"""cdefs for message.py"""

import cython

from ._private.marshaller cimport Marshaller
from .signature cimport Variant, SignatureTree, SignatureType

cdef object ErrorType
cdef object MessageType

cdef object HEADER_PATH
cdef object HEADER_INTERFACE
cdef object HEADER_MEMBER
cdef object HEADER_ERROR_NAME
cdef object HEADER_REPLY_SERIAL
cdef object HEADER_DESTINATION
cdef object HEADER_SENDER
cdef object HEADER_SIGNATURE
cdef object HEADER_UNIX_FDS


cdef object LITTLE_ENDIAN
cdef object PROTOCOL_VERSION

cdef object MESSAGE_FLAG
cdef object MESSAGE_FLAG_NONE
cdef object MESSAGE_TYPE_METHOD_CALL

cdef SignatureTree SIGNATURE_TREE_G
cdef SignatureTree SIGNATURE_TREE_O
cdef SignatureTree SIGNATURE_TREE_S
cdef SignatureTree SIGNATURE_TREE_U

cdef get_signature_tree

cdef class Message:

    cdef public object destination
    cdef public object path
    cdef public object interface
    cdef public object member
    cdef public object message_type
    cdef public object flags
    cdef public object error_name
    cdef public unsigned int reply_serial
    cdef public object sender
    cdef public list unix_fds
    cdef public str signature
    cdef public SignatureTree signature_tree
    cdef public object body
    cdef public unsigned int serial

    @cython.locals(
        body_buffer=bytearray,
        header_buffer=bytearray,
        var=Variant
    )
    cpdef _marshall(self, bint negotiate_unix_fd)

    cdef _fast_init(
        self,
        object destination,
        object path,
        object interface,
        object member,
        object message_type,
        object flags,
        object error_name,
        unsigned int reply_serial,
        object sender,
        list unix_fds,
        SignatureTree signature_tree,
        object body,
        unsigned int serial,
        bint validate
    )



import io
import logging
import traceback
from collections.abc import Callable

from .. import introspection as intr
from .._private.unmarshaller import Unmarshaller
from ..auth import _MAX_AUTH_LINE, Authenticator, AuthExternal
from ..constants import (
    BusType,
    MessageFlag,
    MessageType,
    NameFlag,
    ReleaseNameReply,
    RequestNameReply,
)
from ..errors import AuthError
from ..message import Message
from ..message_bus import BaseMessageBus
from .proxy_object import ProxyObject

_LOGGER = logging.getLogger(__name__)


# glib is optional
_import_error = None
try:
    from gi.repository import GLib

    _GLibSource = GLib.Source
except ImportError as e:
    _import_error = e

    class _GLibSource:
        pass


class _MessageSource(_GLibSource):
    def __init__(self, bus):
        self.unmarshaller = None
        self.bus = bus

    def prepare(self):
        return (False, -1)

    def check(self):
        return False

    def dispatch(self, callback, user_data):
        try:
            while self.bus._stream.readable():
                if not self.unmarshaller:
                    self.unmarshaller = Unmarshaller(self.bus._stream)

                message = self.unmarshaller.unmarshall()
                if message:
                    callback(message)
                    self.unmarshaller = None
                else:
                    break
        except Exception as e:
            self.bus.disconnect()
            self.bus._finalize(e)
            return GLib.SOURCE_REMOVE

        return GLib.SOURCE_CONTINUE


class _MessageWritableSource(_GLibSource):
    def __init__(self, bus):
        self.bus = bus
        self.buf = b""
        self.message_stream = None
        self.chunk_size = 128

    def prepare(self):
        return (False, -1)

    def check(self):
        return False

    def dispatch(self, callback, user_data):
        try:
            if self.buf:
                self.bus._stream.write(self.buf)
                self.buf = b""

            if self.message_stream:
                while True:
                    self.buf = self.message_stream.read(self.chunk_size)
                    if self.buf == b"":
                        break
                    self.bus._stream.write(self.buf)
                    if len(self.buf) < self.chunk_size:
                        self.buf = b""
                        break
                    self.buf = b""

            self.bus._stream.flush()

            if not self.bus._buffered_messages:
                return GLib.SOURCE_REMOVE
            message = self.bus._buffered_messages.pop(0)
            self.message_stream = io.BytesIO(message._marshall(False))
            return GLib.SOURCE_CONTINUE
        except BlockingIOError:
            return GLib.SOURCE_CONTINUE
        except Exception as e:
            self.bus._finalize(e)
            return GLib.SOURCE_REMOVE


class _AuthLineSource(_GLibSource):
    def __init__(self, stream):
        self.stream = stream
        self.buf = b""

    def prepare(self):
        return (False, -1)

    def check(self):
        return False

    def dispatch(self, callback, user_data):
        chunk = self.stream.read()
        if chunk == b"":
            # EOF before CRLF — peer hung up mid-handshake.
            callback(AuthError("connection closed during authentication"))
            return GLib.SOURCE_REMOVE
        if chunk:
            self.buf += chunk
            if len(self.buf) > _MAX_AUTH_LINE:
                callback(AuthError("auth line exceeded maximum size"))
                return GLib.SOURCE_REMOVE
        if self.buf[-2:] == b"\r\n":
            resp = callback(self.buf.decode()[:-2])
            if resp:
                return GLib.SOURCE_REMOVE

        return GLib.SOURCE_CONTINUE


class MessageBus(BaseMessageBus):
    """The message bus implementation for use with the GLib main loop.

    The message bus class is the entry point into all the features of the
    library. It sets up a connection to the DBus daemon and exposes an
    interface to send and receive messages and expose services.

    You must call :func:`connect() <dbus_fast.glib.MessageBus.connect>` or
    :func:`connect_sync() <dbus_fast.glib.MessageBus.connect_sync>` before
    using this message bus.

    :param bus_type: The type of bus to connect to. Affects the search path for
        the bus address.
    :type bus_type: :class:`BusType <dbus_fast.BusType>`
    :param bus_address: A specific bus address to connect to. Should not be
        used under normal circumstances.
    :param auth: The authenticator to use, defaults to an instance of
        :class:`AuthExternal <dbus_fast.auth.AuthExternal>`.
    :type auth: :class:`Authenticator <dbus_fast.auth.Authenticator>`

    :ivar connected: True if this message bus is expected to be able to send
        and receive messages.
    :vartype connected: bool
    :ivar unique_name: The unique name of the message bus connection. It will
        be :class:`None` until the message bus connects.
    :vartype unique_name: str
    """

    def __init__(
        self,
        bus_address: str | None = None,
        bus_type: BusType = BusType.SESSION,
        auth: Authenticator | None = None,
    ):
        if _import_error:
            raise _import_error

        super().__init__(bus_address, bus_type, ProxyObject)
        self._main_context = GLib.main_context_default()
        # buffer messages until connect
        self._buffered_messages = []

        if auth is None:
            self._auth = AuthExternal()
        else:
            self._auth = auth

    def _on_message(self, msg: Message) -> None:
        try:
            self._process_message(msg)
        except Exception as e:
            _LOGGER.exception(
                f"got unexpected error processing a message: {e}.\n{traceback.format_exc()}"
            )

    def connect(
        self,
        connect_notify: None
        | (Callable[["MessageBus", Exception | None], None]) = None,
    ):
        """Connect this message bus to the DBus daemon.

        This method or the synchronous version must be called before the
        message bus can be used.

        :param connect_notify: A callback that will be called with this message
            bus. May return an :class:`Exception` on connection errors or
            :class:`AuthError <dbus_fast.AuthError>` on authorization errors.
        :type callback: :class:`Callable`
        """

        def authenticate_notify(exc):
            if exc is not None:
                if connect_notify is not None:
                    connect_notify(None, exc)
                return
            self.message_source = _MessageSource(self)
            self.message_source.set_callback(self._on_message)
            self.message_source.attach(self._main_context)

            self.writable_source = None

            self.message_source.add_unix_fd(self._fd, GLib.IO_IN)

            def on_hello(reply, err):
                if err:
                    if connect_notify:
                        connect_notify(reply, err)
                    return

                self.unique_name = reply.body[0]

                for m in self._buffered_messages:
                    self.send(m)

                if connect_notify:
                    connect_notify(self, err)

            hello_msg = Message(
                destination="org.freedesktop.DBus",
                path="/org/freedesktop/DBus",
                interface="org.freedesktop.DBus",
                member="Hello",
                serial=self.next_serial(),
            )

            self._method_return_handlers[hello_msg.serial] = on_hello
            self._stream.write(hello_msg._marshall(False))
            self._stream.flush()

        self._authenticate(authenticate_notify)

    def connect_sync(self) -> "MessageBus":
        """Connect this message bus to the DBus daemon.

        This method or the asynchronous version must be called before the
        message bus can be used.

        :returns: This message bus for convenience.
        :rtype: :class:`MessageBus <dbus_fast.glib.MessageBus>`

        :raises:
            - :class:`AuthError <dbus_fast.AuthError>` - If authorization to \
              the DBus daemon failed.
            - :class:`Exception` - If there was a connection error.
        """
        main = GLib.MainLoop()
        connection_error = None

        def connect_notify(bus, err):
            nonlocal connection_error
            connection_error = err
            main.quit()

        self.connect(connect_notify)
        main.run()

        if connection_error:
            raise connection_error

        return self

    def call(
        self,
        msg: Message,
        reply_notify: None
        | (Callable[[Message | None, Exception | None], None]) = None,
    ):
        """Send a method call and asynchronously wait for a reply from the DBus
        daemon.

        :param msg: The method call message to send.
        :type msg: :class:`Message <dbus_fast.Message>`
        :param reply_notify: A callback that will be called with the reply to
            this message. May return an :class:`Exception` on connection errors.
        :type reply_notify: Callable
        """
        BaseMessageBus._check_callback_type(reply_notify)
        self._call(msg, reply_notify)

    def call_sync(self, msg: Message) -> Message | None:
        """Send a method call and synchronously wait for a reply from the DBus
        daemon.

        :param msg: The method call message to send.
        :type msg: :class:`Message <dbus_fast.Message>`

        :returns: A message in reply to the message sent. If the message does
            not expect a reply based on the message flags or type, returns
            ``None`` immediately.
        :rtype: :class:`Message <dbus_fast.Message>`

        :raises:
            - :class:`DBusError <dbus_fast.DBusError>` - If the service threw \
                  an error for the method call or returned an invalid result.
            - :class:`Exception` - If a connection error occurred.
        """
        if (
            msg.flags & MessageFlag.NO_REPLY_EXPECTED
            or msg.message_type is not MessageType.METHOD_CALL
        ):
            self.send(msg)
            return None

        if not msg.serial:
            msg.serial = self.next_serial()

        main = GLib.MainLoop()
        handler_reply = None
        connection_error = None

        def reply_handler(reply, err):
            nonlocal handler_reply
            nonlocal connection_error

            handler_reply = reply
            connection_error = err

            main.quit()

        self._method_return_handlers[msg.serial] = reply_handler
        self.send(msg)
        main.run()

        if connection_error:
            raise connection_error

        return handler_reply

    def introspect_sync(self, bus_name: str, path: str) -> intr.Node:
        """Get introspection data for the node at the given path from the given
        bus name.

        Calls the standard ``org.freedesktop.DBus.Introspectable.Introspect``
        on the bus for the path.

        :param bus_name: The name to introspect.
        :type bus_name: str
        :param path: The path to introspect.
        :type path: str

        :returns: The introspection data for the name at the path.
        :rtype: :class:`Node <dbus_fast.introspection.Node>`

        :raises:
            - :class:`InvalidObjectPathError <dbus_fast.InvalidObjectPathError>` \
                    - If the given object path is not valid.
            - :class:`InvalidBusNameError <dbus_fast.InvalidBusNameError>` - If \
                  the given bus name is not valid.
            - :class:`DBusError <dbus_fast.DBusError>` - If the service threw \
                  an error for the method call or returned an invalid result.
            - :class:`Exception` - If a connection error occurred.
        """
        main = GLib.MainLoop()
        request_result = None
        request_error = None

        def reply_notify(result, err):
            nonlocal request_result
            nonlocal request_error

            request_result = result
            request_error = err

            main.quit()

        super().introspect(bus_name, path, reply_notify)
        main.run()

        if request_error:
            raise request_error

        return request_result

    def request_name_sync(
        self, name: str, flags: NameFlag = NameFlag.NONE
    ) -> RequestNameReply:
        """Request that this message bus owns the given name.

        :param name: The name to request.
        :type name: str
        :param flags: Name flags that affect the behavior of the name request.
        :type flags: :class:`NameFlag <dbus_fast.NameFlag>`

        :returns: The reply to the name request.
        :rtype: :class:`RequestNameReply <dbus_fast.RequestNameReply>`

        :raises:
            - :class:`InvalidBusNameError <dbus_fast.InvalidBusNameError>` - If \
                  the given bus name is not valid.
            - :class:`DBusError <dbus_fast.DBusError>` - If the service threw \
                  an error for the method call or returned an invalid result.
            - :class:`Exception` - If a connection error occurred.
        """
        main = GLib.MainLoop()
        request_result = None
        request_error = None

        def reply_notify(result, err):
            nonlocal request_result
            nonlocal request_error

            request_result = result
            request_error = err

            main.quit()

        super().request_name(name, flags, reply_notify)
        main.run()

        if request_error:
            raise request_error

        return request_result

    def release_name_sync(self, name: str) -> ReleaseNameReply:
        """Request that this message bus release the given name.

        :param name: The name to release.
        :type name: str

        :returns: The reply to the release request.
        :rtype: :class:`ReleaseNameReply <dbus_fast.ReleaseNameReply>`

        :raises:
            - :class:`InvalidBusNameError <dbus_fast.InvalidBusNameError>` - If \
                  the given bus name is not valid.
            - :class:`DBusError <dbus_fast.DBusError>` - If the service threw \
                  an error for the method call or returned an invalid result.
            - :class:`Exception` - If a connection error occurred.
        """
        main = GLib.MainLoop()
        release_result = None
        release_error = None

        def reply_notify(result, err):
            nonlocal release_result
            nonlocal release_error

            release_result = result
            release_error = err

            main.quit()

        super().release_name(name, reply_notify)
        main.run()

        if release_error:
            raise release_error

        return release_result

    def send(self, msg: Message):
        if not msg.serial:
            msg.serial = self.next_serial()

        self._buffered_messages.append(msg)

        if self.unique_name:
            self._schedule_write()

    def get_proxy_object(
        self, bus_name: str, path: str, introspection: intr.Node
    ) -> ProxyObject:
        return super().get_proxy_object(bus_name, path, introspection)

    def _schedule_write(self):
        if self.writable_source is None or self.writable_source.is_destroyed():
            self.writable_source = _MessageWritableSource(self)
            self.writable_source.attach(self._main_context)
            self.writable_source.add_unix_fd(self._fd, GLib.IO_OUT)

    def _authenticate(self, authenticate_notify):
        self._stream.write(b"\0")
        first_line = self._auth._authentication_start()
        if first_line is not None:
            if type(first_line) is not str:
                raise AuthError("authenticator gave response not type str")
            self._stream.write(f"{first_line}\r\n".encode())
            self._stream.flush()

        def line_notify(line):
            try:
                if isinstance(line, Exception):
                    raise line
                resp = self._auth._receive_line(line)
                self._stream.write(Authenticator._format_line(resp))
                self._stream.flush()
                if resp == "BEGIN":
                    self._readline_source = None
                    authenticate_notify(None)
                    return True
            except Exception as e:
                authenticate_notify(e)
                return True

        readline_source = _AuthLineSource(self._stream)
        readline_source.set_callback(line_notify)
        readline_source.add_unix_fd(self._fd, GLib.IO_IN)
        readline_source.attach(self._main_context)
        # make sure it doesnt get cleaned up
        self._readline_source = readline_source



# cython: freethreading_compatible = True

from typing import Any

from ._private.constants import LITTLE_ENDIAN, PROTOCOL_VERSION, HeaderField
from ._private.marshaller import Marshaller
from .constants import ErrorType, MessageFlag, MessageType
from .errors import InvalidMessageError
from .signature import SignatureTree, Variant, get_signature_tree
from .validators import (
    assert_bus_name_valid,
    assert_interface_name_valid,
    assert_member_name_valid,
    assert_object_path_valid,
)

REQUIRED_FIELDS = {
    MessageType.METHOD_CALL.value: ("path", "member"),
    MessageType.SIGNAL.value: ("path", "member", "interface"),
    MessageType.ERROR.value: ("error_name", "reply_serial"),
    MessageType.METHOD_RETURN.value: ("reply_serial",),
}

HEADER_PATH = HeaderField.PATH.value
HEADER_INTERFACE = HeaderField.INTERFACE.value
HEADER_MEMBER = HeaderField.MEMBER.value
HEADER_ERROR_NAME = HeaderField.ERROR_NAME.value
HEADER_REPLY_SERIAL = HeaderField.REPLY_SERIAL.value
HEADER_DESTINATION = HeaderField.DESTINATION.value
HEADER_SIGNATURE = HeaderField.SIGNATURE.value
HEADER_UNIX_FDS = HeaderField.UNIX_FDS.value

MESSAGE_FLAG = MessageFlag

MESSAGE_FLAG_NONE = MessageFlag.NONE
MESSAGE_TYPE_METHOD_CALL = MessageType.METHOD_CALL

SIGNATURE_TREE_G = get_signature_tree("g")
SIGNATURE_TREE_O = get_signature_tree("o")
SIGNATURE_TREE_S = get_signature_tree("s")
SIGNATURE_TREE_U = get_signature_tree("u")

_int = int
_str = str
_bool = bool
_MessageType = MessageType
_MessageFlag = MessageFlag
_list = list


class Message:
    """A class for sending and receiving messages through the
    :class:`MessageBus <dbus_fast.message_bus.BaseMessageBus>` with the
    low-level api.

    A ``Message`` can be constructed by the user to send over the message bus.
    When messages are received, such as from method calls or signal emissions,
    they will use this class as well.

    :ivar destination: The address of the client for which this message is intended.
    :vartype destination: str
    :ivar path: The intended object path exported on the destination bus.
    :vartype path: str
    :ivar interface: The intended interface on the object path.
    :vartype interface: str
    :ivar member: The intended member on the interface.
    :vartype member: str
    :ivar message_type: The type of this message. A method call, signal, method return, or error.
    :vartype message_type: :class:`MessageType`
    :ivar flags: Flags that affect the behavior of this message.
    :vartype flags: :class:`MessageFlag`
    :ivar error_name: If this message is an error, the name of this error. Must be a valid interface name.
    :vartype error_name: str
    :ivar reply_serial: If this is a return type, the serial this message is in reply to.
    :vartype reply_serial: int
    :ivar sender: The address of the sender of this message. Will be a unique name.
    :vartype sender: str
    :ivar unix_fds: A list of unix fds that were sent in the header of this message.
    :vartype unix_fds: list(int)
    :ivar signature: The signature of the body of this message.
    :vartype signature: str
    :ivar signature_tree: The signature parsed as a signature tree.
    :vartype signature_tree: :class:`SignatureTree`
    :ivar body: The body of this message. Must match the signature.
    :vartype body: list(Any)
    :ivar serial: The serial of the message. Will be automatically set during message sending if not present. Use the ``new_serial()`` method of the bus to generate a serial.
    :vartype serial: int

    :raises:
        - :class:`InvalidMessageError` - If the message is malformed or missing fields for the message type.
        - :class:`InvalidSignatureError` - If the given signature is not valid.
        - :class:`InvalidObjectPathError` - If ``path`` is not a valid object path.
        - :class:`InvalidBusNameError` - If ``destination`` is not a valid bus name.
        - :class:`InvalidMemberNameError` - If ``member`` is not a valid member name.
        - :class:`InvalidInterfaceNameError` - If ``error_name`` or ``interface`` is not a valid interface name.
    """

    __slots__ = (
        "body",
        "destination",
        "error_name",
        "flags",
        "interface",
        "member",
        "message_type",
        "path",
        "reply_serial",
        "sender",
        "serial",
        "signature",
        "signature_tree",
        "unix_fds",
    )

    def __init__(
        self,
        destination: str | None = None,
        path: str | None = None,
        interface: str | None = None,
        member: str | None = None,
        message_type: MessageType = MESSAGE_TYPE_METHOD_CALL,
        flags: MessageFlag | int = MESSAGE_FLAG_NONE,
        error_name: str | ErrorType | None = None,
        reply_serial: int | None = None,
        sender: str | None = None,
        unix_fds: list[int] = [],
        signature: SignatureTree | str | None = None,
        body: list[Any] = [],
        serial: int | None = None,
        validate: bool = True,
    ) -> None:
        self._fast_init(
            destination,
            path,
            interface,
            member,
            message_type,
            flags if type(flags) is MESSAGE_FLAG else MESSAGE_FLAG(flags),
            str(error_name.value) if type(error_name) is ErrorType else error_name,
            reply_serial or 0,
            sender,
            unix_fds,
            signature
            if type(signature) is SignatureTree
            else get_signature_tree(signature or ""),
            body,
            serial or 0,
            validate,
        )

    def _fast_init(
        self,
        destination: _str | None,
        path: _str | None,
        interface: _str | None,
        member: _str | None,
        message_type: _MessageType,
        flags: _MessageFlag,
        error_name: _str | None,
        reply_serial: _int,
        sender: _str,
        unix_fds: _list[int],
        signature_tree: SignatureTree,
        body: _list[Any],
        serial: _int,
        validate: _bool,
    ) -> None:
        self.destination = destination
        self.path = path
        self.interface = interface
        self.member = member
        self.message_type = message_type
        self.flags = flags
        self.error_name = error_name
        self.reply_serial = reply_serial
        self.sender = sender
        self.unix_fds = unix_fds
        self.signature = signature_tree.signature
        self.signature_tree = signature_tree
        self.body = body
        self.serial = serial

        if not validate:
            return
        if self.destination is not None:
            assert_bus_name_valid(self.destination)
        if self.interface is not None:
            assert_interface_name_valid(self.interface)
        if self.path is not None:
            assert_object_path_valid(self.path)
        if self.member is not None:
            assert_member_name_valid(self.member)
        if self.error_name is not None:
            assert_interface_name_valid(self.error_name)  # type: ignore[arg-type]

        required_fields = REQUIRED_FIELDS.get(self.message_type.value)
        if not required_fields:
            raise InvalidMessageError(f"got unknown message type: {self.message_type}")
        for field in required_fields:
            if not getattr(self, field):
                raise InvalidMessageError(f"missing required field: {field}")

    def __repr__(self) -> str:
        """Return a string representation of this message."""
        return (
            f"<Message {self.message_type.name} "
            f"serial={self.serial} "
            f"reply_serial={self.reply_serial} "
            f"sender={self.sender} "
            f"destination={self.destination} "
            f"path={self.path} "
            f"interface={self.interface} "
            f"member={self.member} "
            f"error_name={self.error_name} "
            f"signature={self.signature} "
            f"body={self.body}>"
        )

    @staticmethod
    def new_error(
        msg: "Message", error_name: str | ErrorType, error_text: str
    ) -> "Message":
        """A convenience constructor to create an error message in reply to the given message.

        :param msg: The message this error is in reply to.
        :type msg: :class:`Message`
        :param error_name: The name of this error. Must be a valid interface name.
        :type error_name: str
        :param error_text: Human-readable text for the error.

        :returns: The error message.
        :rtype: :class:`Message`

        :raises:
            - :class:`InvalidInterfaceNameError` - If the error_name is not a valid interface name.
        """
        return Message(
            message_type=MessageType.ERROR,
            reply_serial=msg.serial,
            destination=msg.sender,
            error_name=error_name,
            signature="s",
            body=[error_text],
        )

    @staticmethod
    def new_method_return(
        msg: "Message",
        signature: str = "",
        body: list[Any] = [],
        unix_fds: list[int] = [],
    ) -> "Message":
        """A convenience constructor to create a method return to the given method call message.

        :param msg: The method call message this is a reply to.
        :type msg: :class:`Message`
        :param signature: The signature for the message body.
        :type signature: str
        :param body: The body of this message. Must match the signature.
        :type body: list(Any)
        :param unix_fds: List integer file descriptors to send with this message.
        :type unix_fds: list(int)

        :returns: The method return message
        :rtype: :class:`Message`

        :raises:
            - :class:`InvalidSignatureError` - If the signature is not a valid signature.
        """
        return Message(
            message_type=MessageType.METHOD_RETURN,
            reply_serial=msg.serial,
            destination=msg.sender,
            signature=signature,
            body=body,
            unix_fds=unix_fds,
        )

    @staticmethod
    def new_signal(
        path: str,
        interface: str,
        member: str,
        signature: str = "",
        body: list[Any] | None = None,
        unix_fds: list[int] | None = None,
    ) -> "Message":
        """A convenience constructor to create a new signal message.

        :param path: The path of this signal.
        :type path: str
        :param interface: The interface of this signal.
        :type interface: str
        :param member: The member name of this signal.
        :type member: str
        :param signature: The signature of the signal body.
        :type signature: str
        :param body: The body of this signal message.
        :type body: list(Any)
        :param unix_fds: List integer file descriptors to send with this message.
        :type unix_fds: list(int)

        :returns: The signal message.
        :rtype: :class:`Message`

        :raises:
            - :class:`InvalidSignatureError` - If the signature is not a valid signature.
            - :class:`InvalidObjectPathError` - If ``path`` is not a valid object path.
            - :class:`InvalidInterfaceNameError` - If ``interface`` is not a valid interface name.
            - :class:`InvalidMemberNameError` - If ``member`` is not a valid member name.
        """
        return Message(
            message_type=MessageType.SIGNAL,
            interface=interface,
            path=path,
            member=member,
            signature=signature,
            body=body or [],
            unix_fds=unix_fds or [],
        )

    def _marshall(self, negotiate_unix_fd: bool) -> bytearray:
        """Marshall this message into a byte array."""
        # TODO maximum message size is 134217728 (128 MiB)
        body_block = Marshaller(self.signature, self.body)
        body_buffer = body_block._marshall()

        fields = []

        # No verify here since the marshaller will raise an exception if the
        # Variant is invalid.

        if self.path:
            fields.append((HEADER_PATH, Variant._factory(SIGNATURE_TREE_O, self.path)))
        if self.interface:
            fields.append(
                (HEADER_INTERFACE, Variant._factory(SIGNATURE_TREE_S, self.interface))
            )
        if self.member:
            fields.append(
                (HEADER_MEMBER, Variant._factory(SIGNATURE_TREE_S, self.member))
            )
        if self.error_name:
            fields.append(
                (
                    HEADER_ERROR_NAME,
                    Variant._factory(SIGNATURE_TREE_S, self.error_name),
                )
            )
        if self.reply_serial:
            fields.append(
                (
                    HEADER_REPLY_SERIAL,
                    Variant._factory(SIGNATURE_TREE_U, self.reply_serial),
                )
            )
        if self.destination:
            fields.append(
                (
                    HEADER_DESTINATION,
                    Variant._factory(SIGNATURE_TREE_S, self.destination),
                )
            )
        if self.signature:
            fields.append(
                (HEADER_SIGNATURE, Variant._factory(SIGNATURE_TREE_G, self.signature))
            )
        if self.unix_fds and negotiate_unix_fd:
            fields.append(
                (
                    HEADER_UNIX_FDS,
                    Variant._factory(SIGNATURE_TREE_U, len(self.unix_fds)),
                )
            )

        header_body = [
            LITTLE_ENDIAN,
            self.message_type.value,
            self.flags.value,
            PROTOCOL_VERSION,
            len(body_buffer),
            self.serial,
            fields,
        ]
        header_block = Marshaller("yyyyuua(yv)", header_body)
        header_block._marshall()
        header_block._align(8)
        header_buffer = header_block._buffer()
        return header_buffer + body_buffer



__title__ = "dbus_fast"
__description__ = (
    "A performant zero-dependency DBus library for Python with asyncio support"
)
__url__ = "https://github.com/bluetooth-devices/dbus-fast"
__version__ = "4.3.0"
__author__ = "Bluetooth Devices authors, Tony Crisci"
__author_email__ = "bluetooth@koston.org"
__license__ = "MIT"
__copyright__ = "Copyright 2022 Bluetooth Devices authors, 2019 Tony Crisci"



"""Spawn a temporary dbus-daemon for the test session.

On Linux CI the suite is launched under ``dbus-run-session`` so
``DBUS_SESSION_BUS_ADDRESS`` is already set; this fixture is a no-op there.

On macOS (and any other host without a session bus), if ``dbus-daemon`` is
available on ``PATH`` we start one against a temporary unix socket so the
integration tests can connect to a real bus locally.

If ``dbus-daemon`` is not available we yield without setting anything;
integration tests will fail with ``InvalidAddressError`` exactly as they did
before this fixture existed.
"""

from __future__ import annotations

import os
import shutil
import subprocess
import tempfile
from collections.abc import Iterator
from pathlib import Path

import pytest

_SESSION_CONF = """<!DOCTYPE busconfig PUBLIC "-//freedesktop//DTD D-Bus Bus Configuration 1.0//EN"
 "http://www.freedesktop.org/standards/dbus/1.0/busconfig.dtd">
<busconfig>
  <type>session</type>
  <keep_umask/>
  <listen>unix:tmpdir={tmpdir}</listen>
  <auth>EXTERNAL</auth>
  <policy context="default">
    <allow send_destination="*" eavesdrop="true"/>
    <allow eavesdrop="true"/>
    <allow own="*"/>
  </policy>
</busconfig>
"""


@pytest.fixture(scope="session", autouse=True)
def _dbus_session_bus() -> Iterator[None]:
    if os.environ.get("DBUS_SESSION_BUS_ADDRESS"):
        yield
        return

    dbus_daemon = shutil.which("dbus-daemon")
    if dbus_daemon is None:
        yield
        return

    with tempfile.TemporaryDirectory(prefix="dbus-fast-test-") as tmp:
        conf_path = Path(tmp) / "session.conf"
        conf_path.write_text(_SESSION_CONF.format(tmpdir=tmp))

        with subprocess.Popen(  # noqa: S603
            [
                dbus_daemon,
                "--config-file",
                str(conf_path),
                "--print-address",
                "--nofork",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        ) as proc:
            try:
                assert proc.stdout is not None
                address = proc.stdout.readline().strip()
                if not address:
                    stderr = proc.stderr.read() if proc.stderr else ""
                    raise RuntimeError(
                        f"dbus-daemon did not print an address. stderr: {stderr!r}"
                    )

                os.environ["DBUS_SESSION_BUS_ADDRESS"] = address
                try:
                    yield
                finally:
                    os.environ.pop("DBUS_SESSION_BUS_ADDRESS", None)
            finally:
                proc.terminate()
                try:
                    proc.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    proc.kill()
                    proc.wait()



import timeit

from dbus_fast import Message

message = Message(
    destination="org.bluez",
    path="/",
    interface="org.freedesktop.DBus.ObjectManager",
    member="GetManagedObjects",
)


def marhsall_bluez_get_managed_objects_message():
    message._marshall(False)


count = 1000000
time = timeit.Timer(marhsall_bluez_get_managed_objects_message).timeit(count)
print(f"Marshalling {count} bluez get managed objects messages took {time} seconds")



"""Pre-auth DoS guards for the auth-line readers."""

from __future__ import annotations

import asyncio
import socket
from types import SimpleNamespace

import pytest

from dbus_fast.aio.message_bus import MessageBus
from dbus_fast.errors import AuthError


def _fake_aio_self(sock: socket.socket) -> SimpleNamespace:
    return SimpleNamespace(_sock=sock, _loop=asyncio.get_running_loop())


@pytest.mark.asyncio
async def test_auth_readline_raises_on_eof() -> None:
    server, client = socket.socketpair()
    try:
        client.setblocking(False)
        server.close()  # peer EOF before any data is sent

        coro = MessageBus._auth_readline(_fake_aio_self(client))
        with pytest.raises(AuthError):
            await asyncio.wait_for(coro, timeout=1.0)
    finally:
        client.close()


@pytest.mark.asyncio
async def test_auth_readline_rejects_oversize_line() -> None:
    server, client = socket.socketpair()
    try:
        client.setblocking(False)
        server.setblocking(True)
        # Stream junk that never contains \r\n; cap is 16 KiB so 64 KiB is
        # well past the limit but small enough to fit a socket buffer.
        server.sendall(b"A" * 64 * 1024)
        server.close()

        coro = MessageBus._auth_readline(_fake_aio_self(client))
        with pytest.raises(AuthError):
            await asyncio.wait_for(coro, timeout=1.0)
    finally:
        client.close()


@pytest.mark.asyncio
async def test_auth_readline_returns_line() -> None:
    server, client = socket.socketpair()
    try:
        client.setblocking(False)
        server.setblocking(True)
        server.sendall(b"OK 1234\r\n")
        server.close()

        line = await asyncio.wait_for(
            MessageBus._auth_readline(_fake_aio_self(client)), timeout=1.0
        )
        assert line == "OK 1234"
    finally:
        client.close()



"""Tests for the exception hierarchy in ``dbus_fast.errors``.

The hierarchy was introduced to let callers catch every dbus-fast error with
``except DBusFastError`` without giving up the historical ability to catch
specific built-in types like ``ValueError`` or ``TypeError`` for individual
classes (see GH #507).
"""

from __future__ import annotations

import pytest

from dbus_fast import (
    AuthError,
    DBusError,
    DBusFastError,
    InterfaceNotFoundError,
    InternalError,
    InvalidAddressError,
    InvalidBusNameError,
    InvalidInterfaceNameError,
    InvalidIntrospectionError,
    InvalidMemberNameError,
    InvalidMessageError,
    InvalidObjectPathError,
    InvalidSignatureError,
    SignalDisabledError,
    SignatureBodyMismatchError,
)

_ALL_ERRORS = [
    AuthError,
    DBusError,
    InterfaceNotFoundError,
    InternalError,
    InvalidAddressError,
    InvalidBusNameError,
    InvalidInterfaceNameError,
    InvalidIntrospectionError,
    InvalidMemberNameError,
    InvalidMessageError,
    InvalidObjectPathError,
    InvalidSignatureError,
    SignalDisabledError,
    SignatureBodyMismatchError,
]


@pytest.mark.parametrize("err_cls", _ALL_ERRORS)
def test_all_errors_share_dbus_fast_base(err_cls: type[BaseException]) -> None:
    assert issubclass(err_cls, DBusFastError)
    assert issubclass(err_cls, Exception)


@pytest.mark.parametrize(
    "err_cls",
    [
        SignatureBodyMismatchError,
        InvalidSignatureError,
        InvalidAddressError,
        InvalidMessageError,
        InvalidIntrospectionError,
    ],
)
def test_value_error_subclasses_preserved(err_cls: type[BaseException]) -> None:
    assert issubclass(err_cls, ValueError)


@pytest.mark.parametrize(
    "err_cls",
    [
        InvalidBusNameError,
        InvalidObjectPathError,
        InvalidInterfaceNameError,
        InvalidMemberNameError,
    ],
)
def test_type_error_subclasses_preserved(err_cls: type[BaseException]) -> None:
    assert issubclass(err_cls, TypeError)


def test_value_error_subclass_mro_order() -> None:
    """Built-in ``ValueError`` must precede ``DBusFastError`` in the MRO.

    This ordering ensures ``super().__init__`` routes through
    ``ValueError.__init__`` rather than skipping straight to
    ``Exception.__init__``, preserving the behavior of the historical
    single-inheritance ``ValueError`` subclasses.
    """
    mro = InvalidSignatureError.__mro__
    assert mro.index(ValueError) < mro.index(DBusFastError)


def test_type_error_subclass_mro_order() -> None:
    """Built-in ``TypeError`` must precede ``DBusFastError`` in the MRO.

    Same rationale as the ``ValueError`` MRO test: keeps ``super().__init__``
    routing through ``TypeError.__init__`` to match the original
    single-inheritance ``TypeError`` subclasses.
    """
    mro = InvalidBusNameError.__mro__
    assert mro.index(TypeError) < mro.index(DBusFastError)


def test_dbus_fast_error_catches_value_error_subclass() -> None:
    with pytest.raises(DBusFastError):
        raise SignatureBodyMismatchError("body mismatch")


def test_dbus_fast_error_catches_type_error_subclass() -> None:
    with pytest.raises(DBusFastError):
        raise InvalidBusNameError("not.a.valid.bus.name!")


def test_value_error_still_catches_signature_errors() -> None:
    with pytest.raises(ValueError):
        raise InvalidSignatureError("bad signature")


def test_type_error_still_catches_name_errors() -> None:
    with pytest.raises(TypeError):
        raise InvalidMemberNameError("1bad")


def test_invalid_bus_name_error_message_unchanged() -> None:
    err = InvalidBusNameError("bad")
    assert str(err) == "invalid bus name: bad"


def test_invalid_object_path_error_message_unchanged() -> None:
    err = InvalidObjectPathError("nope")
    assert str(err) == "invalid object path: nope"


def test_invalid_interface_name_error_message_unchanged() -> None:
    err = InvalidInterfaceNameError("bad.iface")
    assert str(err) == "invalid interface name: bad.iface"


def test_invalid_member_name_error_message_unchanged() -> None:
    err = InvalidMemberNameError("1bad")
    assert str(err) == "invalid member name: 1bad"


def test_dbus_error_inherits_dbus_fast_error() -> None:
    err = DBusError("org.freedesktop.DBus.Error.Failed", "boom")
    assert isinstance(err, DBusFastError)
    assert err.type == "org.freedesktop.DBus.Error.Failed"
    assert err.text == "boom"


def test_dbus_fast_error_directly_raisable() -> None:
    with pytest.raises(DBusFastError):
        raise DBusFastError("plain base class")


def test_internal_error_is_runtime_error() -> None:
    assert issubclass(InternalError, RuntimeError)


def test_internal_error_is_dbus_fast_error() -> None:
    assert issubclass(InternalError, DBusFastError)


def test_internal_error_mro_order() -> None:
    """``RuntimeError`` must precede ``DBusFastError`` in the MRO.

    Same rationale as the ``ValueError`` / ``TypeError`` MRO tests: keeps
    ``super().__init__`` routing through ``RuntimeError.__init__``.
    """
    mro = InternalError.__mro__
    assert mro.index(RuntimeError) < mro.index(DBusFastError)


def test_runtime_error_still_catches_internal_error() -> None:
    with pytest.raises(RuntimeError):
        raise InternalError("boom")


def test_dbus_fast_error_catches_internal_error() -> None:
    with pytest.raises(DBusFastError):
        raise InternalError("boom")


def test_internal_error_message_unchanged() -> None:
    err = InternalError("internal boom")
    assert str(err) == "internal boom"



"""Build optional cython modules."""

import logging
import os
from distutils.command.build_ext import build_ext

_LOGGER = logging.getLogger(__name__)

try:
    from setuptools import Extension
except ImportError:
    from distutils.core import Extension


TO_CYTHONIZE = [
    "src/dbus_fast/aio/message_reader.py",
    "src/dbus_fast/message.py",
    "src/dbus_fast/message_bus.py",
    "src/dbus_fast/service.py",
    "src/dbus_fast/signature.py",
    "src/dbus_fast/unpack.py",
    "src/dbus_fast/_private/address.py",
    "src/dbus_fast/_private/marshaller.py",
    "src/dbus_fast/_private/unmarshaller.py",
]

EXTENSIONS = [
    Extension(
        ext.removeprefix("src/").removesuffix(".py").replace("/", "."),
        [ext],
        language="c",
        extra_compile_args=["-O3", "-g0"],
    )
    for ext in TO_CYTHONIZE
]


class BuildExt(build_ext):
    def build_extensions(self):
        if self.parallel is None:
            self.parallel = os.cpu_count() or 1
        try:
            super().build_extensions()
        except Exception:
            _LOGGER.debug("Failed to build extensions", exc_info=True)


def build(setup_kwargs):
    if os.environ.get("SKIP_CYTHON"):
        return
    try:
        from Cython.Build import cythonize  # noqa: PLC0415

        setup_kwargs.update(
            {
                "ext_modules": cythonize(
                    EXTENSIONS,
                    compiler_directives={"language_level": "3"},  # Python 3
                ),
                "cmdclass": {"build_ext": BuildExt},
            }
        )
        setup_kwargs["exclude_package_data"] = {
            pkg: ["*.c"] for pkg in setup_kwargs["packages"]
        }
    except Exception:
        if os.environ.get("REQUIRE_CYTHON"):
            raise



"""Pre-auth DoS guards for the GLib auth-line source."""

from __future__ import annotations

import io

import pytest

from dbus_fast.auth import AuthExternal
from dbus_fast.errors import AuthError
from dbus_fast.glib import message_bus as glib_message_bus
from dbus_fast.glib.message_bus import MessageBus, _AuthLineSource
from tests.util import check_gi_repository, skip_reason_no_gi

has_gi = check_gi_repository()

if has_gi:
    from gi.repository import GLib


@pytest.mark.skipif(not has_gi, reason=skip_reason_no_gi)
def test_auth_line_source_rejects_oversize() -> None:
    stream = io.BytesIO(b"A" * 64 * 1024)
    source = _AuthLineSource(stream)

    captured: list[object] = []

    def callback(arg: object) -> bool:
        captured.append(arg)
        return True

    source.dispatch(callback, None)

    assert len(captured) == 1
    assert isinstance(captured[0], AuthError)


@pytest.mark.skipif(not has_gi, reason=skip_reason_no_gi)
def test_auth_line_source_rejects_eof() -> None:
    stream = io.BytesIO(b"")
    source = _AuthLineSource(stream)

    captured: list[object] = []

    def callback(arg: object) -> bool:
        captured.append(arg)
        return True

    source.dispatch(callback, None)

    assert len(captured) == 1
    assert isinstance(captured[0], AuthError)


@pytest.mark.skipif(not has_gi, reason=skip_reason_no_gi)
def test_auth_line_source_passes_complete_line() -> None:
    stream = io.BytesIO(b"OK 1234\r\n")
    source = _AuthLineSource(stream)

    captured: list[object] = []

    def callback(arg: object) -> bool:
        captured.append(arg)
        return True

    source.dispatch(callback, None)

    assert captured == ["OK 1234"]


@pytest.mark.skipif(not has_gi, reason=skip_reason_no_gi)
def test_auth_line_source_continues_when_read_returns_none() -> None:
    # Non-blocking reads can return None when no data is ready; dispatch
    # must keep the source attached rather than treat that as EOF.
    class NoneStream:
        def read(self) -> None:
            return None

    source = _AuthLineSource(NoneStream())

    captured: list[object] = []

    def callback(arg: object) -> bool:
        captured.append(arg)
        return True

    result = source.dispatch(callback, None)

    assert captured == []
    assert result == GLib.SOURCE_CONTINUE


@pytest.mark.skipif(not has_gi, reason=skip_reason_no_gi)
def test_line_notify_forwards_auth_error_to_notify() -> None:
    # When _AuthLineSource hands line_notify an Exception instead of a line
    # (EOF / oversize), line_notify must surface it via authenticate_notify
    # rather than crash.
    bus = MessageBus.__new__(MessageBus)
    bus._auth = AuthExternal()
    bus._fd = -1
    bus._main_context = None

    class FakeStream:
        def __init__(self) -> None:
            self.written = bytearray()

        def write(self, data: bytes) -> int:
            self.written.extend(data)
            return len(data)

        def flush(self) -> None:
            pass

    bus._stream = FakeStream()

    captured_callbacks: list[object] = []

    class CapturingSource:
        def __init__(self, stream: object) -> None:
            self.stream = stream

        def set_callback(self, cb: object) -> None:
            captured_callbacks.append(cb)

        def add_unix_fd(self, fd: int, mask: object) -> None:
            pass

        def attach(self, ctx: object) -> None:
            pass

    notifications: list[object] = []

    def notify(exc: object) -> None:
        notifications.append(exc)

    original = glib_message_bus._AuthLineSource
    glib_message_bus._AuthLineSource = CapturingSource
    try:
        bus._authenticate(notify)
    finally:
        glib_message_bus._AuthLineSource = original

    assert len(captured_callbacks) == 1
    line_notify = captured_callbacks[0]

    err = AuthError("connection closed during authentication")
    result = line_notify(err)

    assert notifications == [err]
    assert result is True



# cython: freethreading_compatible = True

from __future__ import annotations

from collections.abc import Callable
from struct import Struct, error
from typing import Any

from ..errors import InternalError
from ..signature import SignatureType, Variant, get_signature_tree

PACK_LITTLE_ENDIAN = "<"

PACK_UINT32 = Struct(f"{PACK_LITTLE_ENDIAN}I").pack
PACKED_UINT32_ZERO = PACK_UINT32(0)
PACKED_BOOL_FALSE = PACK_UINT32(0)
PACKED_BOOL_TRUE = PACK_UINT32(1)

_int = int
_bytes = bytes
_str = str


class Marshaller:
    """Marshall data for Dbus."""

    __slots__ = ("_buf", "body", "signature_tree")

    def __init__(self, signature: str, body: list[Any]) -> None:
        """Marshaller constructor."""
        self.signature_tree = get_signature_tree(signature)
        self._buf = bytearray()
        self.body = body

    @property
    def buffer(self) -> bytearray:
        return self._buf

    def _buffer(self) -> bytearray:
        return self._buf

    def align(self, n: _int) -> int:
        return self._align(n)

    def _align(self, n: _int) -> _int:
        offset = n - len(self._buf) % n
        if offset == 0 or offset == n:
            return 0
        for _ in range(offset):
            self._buf.append(0)
        return offset

    def write_boolean(self, boolean: bool, type_: SignatureType) -> int:
        return self._write_boolean(boolean)

    def _write_boolean(self, boolean: bool) -> int:
        written = self._align(4)
        self._buf += PACKED_BOOL_TRUE if boolean else PACKED_BOOL_FALSE
        return written + 4

    def write_signature(self, signature: str, type_: SignatureType) -> int:
        return self._write_signature(signature.encode())

    def _write_signature(self, signature_bytes: _bytes) -> int:
        signature_len = len(signature_bytes)
        buf = self._buf
        buf.append(signature_len)
        buf += signature_bytes
        buf.append(0)
        return signature_len + 2

    def write_string(self, value: _str, type_: SignatureType) -> int:
        return self._write_string(value)

    def _write_string(self, value: _str) -> int:
        value_bytes = value.encode()
        value_len = len(value_bytes)
        written = self._align(4) + 4
        buf = self._buf
        buf += PACK_UINT32(value_len)
        buf += value_bytes
        written += value_len
        buf.append(0)
        written += 1
        return written

    def write_variant(self, variant: Variant, type_: SignatureType) -> int:
        return self._write_variant(variant, type_)

    def _write_variant(self, variant: Variant, type_: SignatureType) -> int:
        signature = variant.signature
        signature_bytes = signature.encode()
        written = self._write_signature(signature_bytes)
        written += self._write_single(variant.type, variant.value)
        return written

    def write_array(
        self, array: bytes | list[Any] | dict[Any, Any], type_: SignatureType
    ) -> int:
        return self._write_array(array, type_)

    def _write_array(
        self, array: bytes | list[Any] | dict[Any, Any], type_: SignatureType
    ) -> int:
        # TODO max array size is 64MiB (67108864 bytes)
        written = self._align(4)
        # length placeholder
        buf: bytearray = self._buf
        offset = len(buf)
        written += self._align(4) + 4
        buf += PACKED_UINT32_ZERO
        child_type = type_.children[0]
        token = child_type.token

        if token in "xtd{(":
            # the first alignment is not included in array size
            written += self._align(8)

        array_len = 0
        if token == "{":
            for key, value in array.items():  # type: ignore[union-attr]
                array_len += self.write_dict_entry([key, value], child_type)
        elif token == "y":
            array_len = len(array)
            buf += array  # type: ignore[arg-type]
        elif token == "(":
            for value in array:
                array_len += self._write_struct(value, child_type)
        else:
            writer, packer, size = self._writers[token]
            if not writer:
                for value in array:
                    array_len += self._align(size) + size
                    buf += packer(value)  # type: ignore[misc]
            else:
                for value in array:
                    array_len += writer(self, value, child_type)

        array_len_packed = PACK_UINT32(array_len)
        for i in range(offset, offset + 4):
            buf[i] = array_len_packed[i - offset]

        return written + array_len

    def write_struct(self, array: tuple[Any] | list[Any], type_: SignatureType) -> int:
        return self._write_struct(array, type_)

    def _write_struct(self, array: tuple[Any] | list[Any], type_: SignatureType) -> int:
        written = self._align(8)
        for i, value in enumerate(array):
            written += self._write_single(type_.children[i], value)
        return written

    def write_dict_entry(self, dict_entry: list[Any], type_: SignatureType) -> int:
        written = self._align(8)
        written += self._write_single(type_.children[0], dict_entry[0])
        written += self._write_single(type_.children[1], dict_entry[1])
        return written

    def _write_single(self, type_: SignatureType, body: Any) -> int:
        t = type_.token
        if t == "y":
            self._buf.append(body)
            return 1
        if t == "u":
            written = self._align(4)
            self._buf += PACK_UINT32(body)
            return written + 4
        if t == "a":
            return self._write_array(body, type_)
        if t == "s" or t == "o":
            return self._write_string(body)
        if t == "v":
            return self._write_variant(body, type_)
        if t == "b":
            return self._write_boolean(body)
        writer, packer, size = self._writers[t]
        if not writer:
            written = self._align(size)
            self._buf += packer(body)  # type: ignore[misc]
            return written + size
        return writer(self, body, type_)

    def marshall(self) -> bytearray:
        """Marshalls the body into a byte array"""
        return self._marshall()

    def _marshall(self) -> bytearray:
        """Marshalls the body into a byte array"""
        try:
            return self._construct_buffer()
        except KeyError as ex:
            raise NotImplementedError(
                f'type is not implemented yet: "{ex.args}"'
            ) from ex
        except error:
            self.signature_tree.verify(self.body)
        raise InternalError("should not reach here")  # pragma: no cover

    def _construct_buffer(self) -> bytearray:
        self._buf.clear()
        body = self.body
        for i, type_ in enumerate(self.signature_tree.types):
            self._write_single(type_, body[i])
        return self._buf

    _writers: dict[
        str,
        tuple[
            Callable[[Any, Any, SignatureType], int] | None,
            Callable[[Any], bytes] | None,
            int,
        ],
    ] = {
        "y": (None, Struct(f"{PACK_LITTLE_ENDIAN}B").pack, 1),
        "b": (write_boolean, None, 0),
        "n": (None, Struct(f"{PACK_LITTLE_ENDIAN}h").pack, 2),
        "q": (None, Struct(f"{PACK_LITTLE_ENDIAN}H").pack, 2),
        "i": (None, Struct(f"{PACK_LITTLE_ENDIAN}i").pack, 4),
        "u": (None, PACK_UINT32, 4),
        "x": (None, Struct(f"{PACK_LITTLE_ENDIAN}q").pack, 8),
        "t": (None, Struct(f"{PACK_LITTLE_ENDIAN}Q").pack, 8),
        "d": (None, Struct(f"{PACK_LITTLE_ENDIAN}d").pack, 8),
        "h": (None, Struct(f"{PACK_LITTLE_ENDIAN}I").pack, 4),
        "o": (write_string, None, 0),
        "s": (write_string, None, 0),
        "g": (write_signature, None, 0),
        "a": (write_array, None, 0),
        "(": (write_struct, None, 0),
        "{": (write_dict_entry, None, 0),
        "v": (write_variant, None, 0),
    }



import pytest

from dbus_fast.aio import MessageBus
from dbus_fast.errors import InternalError, InvalidAddressError
from dbus_fast.message_bus import BaseMessageBus


@pytest.mark.asyncio
async def test_tcp_socket_cleanup_on_connect_fail() -> None:
    """Test that socket resources are cleaned up on a failed TCP connection."""

    # A bit ugly, but we need to access members of the class after __init__()
    # raises, so we need to split __new__() and __init__().
    bus = MessageBus.__new__(MessageBus)

    with pytest.raises(ConnectionRefusedError):
        bus.__init__("tcp:host=127.0.0.1,port=1")

    assert bus._stream.closed
    assert bus._sock._closed


@pytest.mark.asyncio
async def test_unix_socket_cleanup_on_connect_fail() -> None:
    """Test that socket resources are cleaned up on a failed Unix socket connection."""

    # A bit ugly, but we need to access members of the class after __init__()
    # raises, so we need to split __new__() and __init__().
    bus = MessageBus.__new__(MessageBus)

    with pytest.raises(FileNotFoundError):
        bus.__init__("unix:path=/there-is-no-way-that-this-file-should-exist")

    assert bus._stream.closed
    assert bus._sock._closed


@pytest.mark.asyncio
async def test_tcp_socket_cleanup_with_host_only() -> None:
    """Test TCP connection with host option only (no port)."""
    bus = MessageBus.__new__(MessageBus)

    with pytest.raises(OSError):
        # Port defaults to 0, which will fail
        bus.__init__("tcp:host=127.0.0.1")

    assert bus._stream.closed
    assert bus._sock._closed


@pytest.mark.asyncio
async def test_tcp_socket_cleanup_with_port_only() -> None:
    """Test TCP connection with port option only (no host)."""
    bus = MessageBus.__new__(MessageBus)

    with pytest.raises(OSError):
        # Host defaults to empty string, which will fail
        bus.__init__("tcp:port=1")

    assert bus._stream.closed
    assert bus._sock._closed


@pytest.mark.asyncio
async def test_unix_socket_abstract_cleanup_on_connect_fail() -> None:
    """Test that socket resources are cleaned up on a failed abstract Unix socket connection."""
    bus = MessageBus.__new__(MessageBus)

    # On Linux: ConnectionRefusedError, on macOS: FileNotFoundError
    with pytest.raises((FileNotFoundError, ConnectionRefusedError)):
        bus.__init__("unix:abstract=/tmp/nonexistent-abstract-socket")

    assert bus._stream.closed
    assert bus._sock._closed


@pytest.mark.asyncio
async def test_unix_socket_invalid_path_specifier() -> None:
    """Test that Unix socket with invalid path specifier raises error."""

    with pytest.raises(
        InvalidAddressError, match="got unix transport with unknown path specifier"
    ):
        MessageBus("unix:invalid=foo")


@pytest.mark.asyncio
async def test_unknown_socket_type() -> None:
    """Test that unknown socket types raise InvalidAddressError."""

    with pytest.raises(InvalidAddressError, match="got unknown address transport"):
        MessageBus("unknown:works=nope")


def test_get_proxy_object_raises_when_proxy_object_class_missing() -> None:
    bus = BaseMessageBus.__new__(BaseMessageBus)
    bus._ProxyObject = None
    with pytest.raises(InternalError):
        bus.get_proxy_object("com.example.Test", "/com/example/Test", "<node/>")



"""
The :mod:`dbus_fast.annotations` module contains :class:`DBusSignature` that
can be used with :class:`typing.Annotated` to provide D-Bus signature annotations
in a manner that is compatible with type hints. This applies to methods decorated
with :meth:`dbus_property <dbus_fast.service.dbus_property>`,
:meth:`dbus_method <dbus_fast.service.dbus_method>` or :meth:`dbus_signal
<dbus_fast.service.dbus_signal>`.

The module also includes some type aliases for common D-Bus types. You can
construct your own annotated types like this::

    from typing import Annotated

    MyDBusStruct = Annotated[tuple[int, str], DBusSignature("(is)")]

Or you can use ``Annotated[...]`` directly without creating an alias.

.. versionadded:: v4.0.0
    Prior to this version, annotations had to be specified as D-Bus signature
    strings.
"""

from dataclasses import dataclass
from typing import Annotated

from dbus_fast.signature import Variant

__all__ = [
    "DBusBool",
    "DBusByte",
    "DBusBytes",
    "DBusDict",
    "DBusDouble",
    "DBusInt16",
    "DBusInt32",
    "DBusInt64",
    "DBusObjectPath",
    "DBusSignature",
    "DBusSignatureType",
    "DBusStr",
    "DBusUInt16",
    "DBusUInt32",
    "DBusUInt64",
    "DBusUnixFd",
    "DBusVariant",
]


@dataclass(frozen=True, slots=True)
class DBusSignature:
    """
    A D-Bus signature annotation.

    This class is used to create D-Bus type annotations. For example::

        from typing import Annotated
        from dbus_fast.annotations import DBusSignature

        MyDBusStruct = Annotated[tuple[int, str], DBusSignature("(is)")]
    """

    signature: str


DBusBool = Annotated[bool, DBusSignature("b")]
"""
D-Bus BOOLEAN type ("b").
"""
DBusByte = Annotated[int, DBusSignature("y")]
"""
D-Bus BYTE type ("y").
"""
DBusInt16 = Annotated[int, DBusSignature("n")]
"""
D-Bus INT16 type ("n").
"""
DBusUInt16 = Annotated[int, DBusSignature("q")]
"""
D-Bus UINT16 type ("q").
"""
DBusInt32 = Annotated[int, DBusSignature("i")]
"""
D-Bus INT32 type ("i").
"""
DBusUInt32 = Annotated[int, DBusSignature("u")]
"""
D-Bus UINT32 type ("u").
"""
DBusInt64 = Annotated[int, DBusSignature("x")]
"""
D-Bus INT64 type ("x").
"""
DBusUInt64 = Annotated[int, DBusSignature("t")]
"""
D-Bus UINT64 type ("t").
"""
DBusDouble = Annotated[float, DBusSignature("d")]
"""
D-Bus DOUBLE type ("d").
"""
DBusStr = Annotated[str, DBusSignature("s")]
"""
D-Bus STRING type ("s").
"""
DBusObjectPath = Annotated[str, DBusSignature("o")]
"""
D-Bus OBJECT_PATH type ("o").
"""
DBusSignatureType = Annotated[str, DBusSignature("g")]
"""
D-Bus SIGNATURE type ("g").

.. note:: This one doesn't follow the established naming convention since
    "DBusSignature" is used for the annotation itself.
"""
DBusVariant = Annotated[Variant, DBusSignature("v")]
"""
D-Bus VARIANT type ("v").
"""
DBusDict = Annotated[dict[str, Variant], DBusSignature("a{sv}")]
"""
D-Bus ARRAY of DICT_ENTRY type with STRING keys and VARIANT values ("a{sv}").
"""
DBusUnixFd = Annotated[int, DBusSignature("h")]
"""
D-Bus UNIX_FD type ("h").
"""
DBusBytes = Annotated[bytes, DBusSignature("ay")]
"""
D-Bus ARRAY of BYTE type ("ay").
"""



import asyncio
import sys

import pytest

from dbus_fast import Message, MessageType, aio, glib
from dbus_fast.annotations import DBusBytes
from dbus_fast.service import ServiceInterface, dbus_method
from tests.util import check_gi_repository, skip_reason_no_gi

has_gi = check_gi_repository()


class ExampleInterface(ServiceInterface):
    def __init__(self):
        super().__init__("example.interface")

    @dbus_method()
    def echo_bytes(self, what: DBusBytes) -> DBusBytes:
        return what


@pytest.mark.asyncio
async def test_aio_big_message():
    "this tests that nonblocking reads and writes actually work for aio"
    bus1 = await aio.MessageBus().connect()
    bus2 = await aio.MessageBus().connect()
    interface = ExampleInterface()
    bus1.export("/test/path", interface)

    # two megabytes
    big_body = [bytes(1000000) * 2]
    result = await bus2.call(
        Message(
            destination=bus1.unique_name,
            path="/test/path",
            interface=interface.name,
            member="echo_bytes",
            signature="ay",
            body=big_body,
        )
    )
    assert result.message_type == MessageType.METHOD_RETURN, result.body[0]
    assert result.body[0] == big_body[0]

    bus1.disconnect()
    bus2.disconnect()
    await asyncio.wait_for(bus1.wait_for_disconnect(), timeout=1)
    await asyncio.wait_for(bus2.wait_for_disconnect(), timeout=1)


@pytest.mark.skipif(not has_gi, reason=skip_reason_no_gi)
@pytest.mark.skipif(
    sys.version_info[:3][1] in (10, 11, 12, 13),
    reason="segfaults on py3.10,py3.11,py3.12,py3.13",
)
def test_glib_big_message():
    "this tests that nonblocking reads and writes actually work for glib"
    bus1 = glib.MessageBus().connect_sync()
    bus2 = glib.MessageBus().connect_sync()
    interface = ExampleInterface()
    bus1.export("/test/path", interface)

    # two megabytes
    big_body = [bytes(1000000) * 2]
    result = bus2.call_sync(
        Message(
            destination=bus1.unique_name,
            path="/test/path",
            interface=interface.name,
            member="echo_bytes",
            signature="ay",
            body=big_body,
        )
    )
    assert result.message_type == MessageType.METHOD_RETURN, result.body[0]
    assert result.body[0] == big_body[0]

    bus1.disconnect()
    bus2.disconnect()



import io
import timeit

from dbus_fast._private.unmarshaller import Unmarshaller

#  cythonize -X language_level=3 -a -i  src/dbus_fast/_private/unmarshaller.py

with open("tests/data/get_managed_objects.hex") as fp:
    msg = fp.read()


stream = io.BytesIO(bytes.fromhex(msg))

unmarshaller = Unmarshaller(stream)


def unmarhsall_bluez_get_managed_objects_message():
    stream.seek(0)
    unmarshaller.unmarshall()


count = 10000
time = timeit.Timer(unmarhsall_bluez_get_managed_objects_message).timeit(count)
print(f"Unmarshalling {count} bluetooth GetManagedObjects messages took {time} seconds")



import asyncio

import pytest

from dbus_fast import ErrorType, Message, MessageFlag, MessageType
from dbus_fast.aio import MessageBus


@pytest.mark.asyncio
async def test_standard_interfaces():
    bus = await MessageBus().connect()
    msg = Message(
        destination="org.freedesktop.DBus",
        path="/org/freedesktop/DBus",
        interface="org.freedesktop.DBus",
        member="ListNames",
        serial=bus.next_serial(),
    )
    reply = await bus.call(msg)

    assert reply.message_type == MessageType.METHOD_RETURN
    assert reply.reply_serial == msg.serial
    assert reply.signature == "as"
    assert bus.unique_name in reply.body[0]

    msg.interface = "org.freedesktop.DBus.Introspectable"
    msg.member = "Introspect"
    msg.serial = bus.next_serial()

    reply = await bus.call(msg)
    assert reply.message_type == MessageType.METHOD_RETURN
    assert reply.reply_serial == msg.serial
    assert reply.signature == "s"
    assert type(reply.body[0]) is str

    msg.member = "MemberDoesNotExist"
    msg.serial = bus.next_serial()

    reply = await bus.call(msg)
    assert reply.message_type == MessageType.ERROR
    assert reply.reply_serial == msg.serial
    assert reply.error_name
    assert reply.signature == "s"
    assert type(reply.body[0]) is str

    bus.disconnect()
    await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_error_handling():
    bus = await MessageBus().connect()
    msg = Message(
        destination="org.freedesktop.DBus",
        path="/org/freedesktop/DBus",
        interface="org.freedesktop.DBus",
        member="InvalidMember",
        serial=bus.next_serial(),
    )
    reply = await bus.call(msg)

    assert reply.message_type == MessageType.ERROR
    assert reply.reply_serial == msg.serial
    assert reply.signature == "s"

    bus.disconnect()
    await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_sending_messages_between_buses():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    msg = Message(
        destination=bus1.unique_name,
        path="/org/test/path",
        interface="org.test.iface",
        member="SomeMember",
        serial=bus2.next_serial(),
    )

    def message_handler(sent):
        if sent.sender == bus2.unique_name and sent.serial == msg.serial:
            assert sent.path == msg.path
            assert sent.serial == msg.serial
            assert sent.interface == msg.interface
            assert sent.member == msg.member
            bus1.send(Message.new_method_return(sent, "s", ["got it"]))
            bus1.remove_message_handler(message_handler)
            return True

    bus1.add_message_handler(message_handler)

    reply = await bus2.call(msg)

    assert reply.message_type == MessageType.METHOD_RETURN
    assert reply.sender == bus1.unique_name
    assert reply.signature == "s"
    assert reply.body == ["got it"]
    assert reply.reply_serial == msg.serial

    def message_handler_error(sent):
        if sent.sender == bus2.unique_name and sent.serial == msg.serial:
            assert sent.path == msg.path
            assert sent.serial == msg.serial
            assert sent.interface == msg.interface
            assert sent.member == msg.member
            bus1.send(Message.new_error(sent, "org.test.Error", "throwing an error"))
            bus1.remove_message_handler(message_handler_error)
            return True

    bus1.add_message_handler(message_handler_error)

    msg.serial = bus2.next_serial()

    reply = await bus2.call(msg)

    assert reply.message_type == MessageType.ERROR
    assert reply.sender == bus1.unique_name
    assert reply.reply_serial == msg.serial
    assert reply.error_name == "org.test.Error"
    assert reply.signature == "s"
    assert reply.body == ["throwing an error"]

    msg.serial = bus2.next_serial()
    msg.flags = MessageFlag.NO_REPLY_EXPECTED

    with pytest.warns(DeprecationWarning):
        reply = await bus2.call(msg)
        assert reply is None

    reply = await bus2.send(msg)
    assert reply is None

    bus1.disconnect()
    bus2.disconnect()
    await asyncio.wait_for(bus1.wait_for_disconnect(), timeout=1)
    await asyncio.wait_for(bus2.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_sending_signals_between_buses():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    add_match_msg = Message(
        destination="org.freedesktop.DBus",
        path="/org/freedesktop/DBus",
        interface="org.freedesktop.DBus",
        member="AddMatch",
        signature="s",
        body=[f"sender={bus2.unique_name}"],
    )

    await bus1.call(add_match_msg)

    async def wait_for_message():
        future = asyncio.get_running_loop().create_future()

        def message_handler(signal):
            if signal.sender == bus2.unique_name:
                bus1.remove_message_handler(message_handler)
                future.set_result(signal)

        bus1.add_message_handler(message_handler)
        return await future

    bus2.send(
        Message.new_signal(
            "/org/test/path", "org.test.interface", "SomeSignal", "s", ["a signal"]
        )
    )

    signal = await wait_for_message()

    assert signal.message_type == MessageType.SIGNAL
    assert signal.path == "/org/test/path"
    assert signal.interface == "org.test.interface"
    assert signal.member == "SomeSignal"
    assert signal.signature == "s"
    assert signal.body == ["a signal"]

    bus1.disconnect()
    bus2.disconnect()
    await asyncio.wait_for(bus1.wait_for_disconnect(), timeout=1)
    await asyncio.wait_for(bus2.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_internal_error_reply_does_not_leak_traceback() -> None:
    """A raising message handler must surface as a sanitized INTERNAL_ERROR."""
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    msg = Message(
        destination=bus1.unique_name,
        path="/org/test/path",
        interface="org.test.iface",
        member="SomeMember",
        serial=bus2.next_serial(),
    )

    def raising_handler(sent):
        if sent.sender == bus2.unique_name and sent.serial == msg.serial:
            bus1.remove_message_handler(raising_handler)
            raise ValueError("must not appear on the wire")

    bus1.add_message_handler(raising_handler)

    reply = await bus2.call(msg)

    assert reply.message_type == MessageType.ERROR
    assert reply.error_name == ErrorType.INTERNAL_ERROR.value
    body = reply.body[0]
    # The body must not leak a Python traceback, file paths, or the
    # exception's str() — only the exception class name is disclosed.
    assert "Traceback" not in body, body
    assert 'File "' not in body, body
    assert "must not appear on the wire" not in body, body
    assert "ValueError" in body, body

    bus1.disconnect()
    bus2.disconnect()
    await asyncio.wait_for(bus1.wait_for_disconnect(), timeout=1)
    await asyncio.wait_for(bus2.wait_for_disconnect(), timeout=1)



"""This tests setting a hardcoded UID in AuthExternal"""

import pytest

from dbus_fast.auth import (
    UID_NOT_SPECIFIED,
    AuthAnnonymous,
    AuthAnonymous,
    AuthExternal,
)
from dbus_fast.errors import AuthError


def test_annonymous_backcompat():
    auth = AuthAnnonymous()
    assert isinstance(auth, AuthAnonymous)


def test_uid_is_set():
    auth = AuthExternal(uid=999)
    assert auth._authentication_start() == "AUTH EXTERNAL 393939"


def test_auth_external_no_uid():
    """Test AuthExternal with UID_NOT_SPECIFIED"""
    auth = AuthExternal(uid=UID_NOT_SPECIFIED)
    assert auth._authentication_start() == "AUTH EXTERNAL"
    assert auth._receive_line("DATA") == "DATA"
    with pytest.raises(AuthError):
        auth._receive_line("REJECTED")



# cython: freethreading_compatible = True

from __future__ import annotations

import os
import re
from urllib.parse import unquote

from ..constants import BusType
from ..errors import InternalError, InvalidAddressError

invalid_address_chars_re = re.compile(r"[^-0-9A-Za-z_/.%]")

str_ = str


def parse_address(address_str: str_) -> list[tuple[str, dict[str, str]]]:
    """Parse a dbus address string into a list of addresses."""
    addresses: list[tuple[str, dict[str, str]]] = []

    for address in address_str.split(";"):
        if not address:
            continue
        if address.find(":") == -1:
            raise InvalidAddressError("address did not contain a transport")

        transport, opt_string = address.split(":", 1)
        options: dict[str, str] = {}

        for kv in opt_string.split(","):
            if not kv:
                continue
            if kv.find("=") == -1:
                raise InvalidAddressError("address option did not contain a value")
            k, v = kv.split("=", 1)
            if invalid_address_chars_re.search(v):
                raise InvalidAddressError("address contains invalid characters")
            # XXX the actual unquote rules are simpler than this
            options[k] = unquote(v)

        addresses.append((transport, options))

    if not addresses:
        raise InvalidAddressError(
            f'address string contained no addresses: "{address_str}"'
        )

    return addresses


def get_system_bus_address() -> str:
    """Get the system bus address from the environment or return the default."""
    return (
        os.environ.get("DBUS_SYSTEM_BUS_ADDRESS")
        or "unix:path=/var/run/dbus/system_bus_socket"
    )


display_re = re.compile(r".*:([0-9]+)\.?.*")
remove_quotes_re = re.compile(r"""^['"]?(.*?)['"]?$""")


def get_session_bus_address() -> str:
    """Get the session bus address from the environment or return the default."""
    dbus_session_bus_address = os.environ.get("DBUS_SESSION_BUS_ADDRESS")
    if dbus_session_bus_address:
        return dbus_session_bus_address

    home = os.environ["HOME"]
    if "DISPLAY" not in os.environ:
        raise InvalidAddressError(
            "DBUS_SESSION_BUS_ADDRESS not set and could not get DISPLAY environment variable to get bus address"
        )

    display = os.environ["DISPLAY"]
    try:
        display = display_re.search(display).group(1)
    except Exception as ex:
        raise InvalidAddressError(
            f"DBUS_SESSION_BUS_ADDRESS not set and could not parse DISPLAY environment variable to get bus address: {display}"
        ) from ex

    # XXX: this will block but they're very small files and fs operations
    # should be fairly reliable. fix this by passing in an async func to read
    # the file for each io backend.
    machine_id = None
    with open("/var/lib/dbus/machine-id") as f:
        machine_id = f.read().rstrip()

    dbus_info_file_name = f"{home}/.dbus/session-bus/{machine_id}-{display}"
    dbus_info: str | None = None
    try:
        with open(dbus_info_file_name) as f:
            dbus_info = f.read().rstrip()
    except Exception as ex:
        raise InvalidAddressError(
            f"could not open dbus info file: {dbus_info_file_name}"
        ) from ex

    for line in dbus_info.split("\n"):
        if line.strip().startswith("DBUS_SESSION_BUS_ADDRESS="):
            _, addr = line.split("=", 1)
            if not addr:
                raise InvalidAddressError(
                    f"DBUS_SESSION_BUS_ADDRESS variable not set correctly in dbus info file: {dbus_info_file_name}"
                )
            addr = remove_quotes_re.search(addr).group(1)
            return addr

    raise InvalidAddressError("could not find dbus session bus address")


def get_bus_address(bus_type: BusType) -> str:
    """Get the address of the bus specified by the bus type."""
    if bus_type == BusType.SESSION:
        return get_session_bus_address()
    if bus_type == BusType.SYSTEM:
        return get_system_bus_address()
    raise InternalError(f"got unknown bus type: {bus_type}")



# NOTE: This is not strictly needed, but included to get code coverage for
# deferred evaluation of annotations. Do not remove this line.
from __future__ import annotations

import asyncio
from typing import Annotated, no_type_check

import pytest

from dbus_fast import (
    DBusError,
    ErrorType,
    Message,
    MessageFlag,
    MessageType,
    SignatureTree,
    Variant,
)
from dbus_fast.aio import MessageBus
from dbus_fast.annotations import DBusDict, DBusSignature, DBusStr, DBusVariant
from dbus_fast.service import ServiceInterface, dbus_method


class ExampleInterface(ServiceInterface):
    def __init__(self, name: str) -> None:
        super().__init__(name)

    @dbus_method()
    def echo(self, what: DBusStr) -> DBusStr:
        assert type(self) is ExampleInterface
        return what

    # This one intentionally keeps string-style annotations for coverage purposes.
    @no_type_check
    @dbus_method()
    def echo_multiple(self, what1: "s", what2: "s") -> "ss":  # noqa: UP037
        assert type(self) is ExampleInterface
        return what1, what2

    @dbus_method()
    def echo_containers(
        self,
        array: Annotated[list[str], DBusSignature("as")],
        variant: DBusVariant,
        dict_entries: DBusDict,
        struct: Annotated[
            tuple[str, tuple[str, tuple[Variant]]], DBusSignature("(s(s(v)))")
        ],
    ) -> Annotated[
        tuple[
            list[str],
            Variant,
            dict[str, Variant],
            tuple[str, tuple[str, tuple[Variant]]],
        ],
        DBusSignature("asva{sv}(s(s(v)))"),
    ]:
        assert type(self) is ExampleInterface
        return array, variant, dict_entries, struct

    @dbus_method()
    def ping(self):
        assert type(self) is ExampleInterface

    @dbus_method(name="renamed")
    def original_name(self):
        assert type(self) is ExampleInterface

    @dbus_method(disabled=True)
    def not_here(self):
        assert type(self) is ExampleInterface

    @dbus_method()
    def throws_unexpected_error(self):
        assert type(self) is ExampleInterface
        raise Exception("oops")

    @dbus_method()
    def throws_dbus_error(self):
        assert type(self) is ExampleInterface
        raise DBusError("test.error", "an error occurred")


class AsyncInterface(ServiceInterface):
    def __init__(self, name: str) -> None:
        super().__init__(name)

    @dbus_method()
    async def echo(self, what: DBusStr) -> DBusStr:
        assert type(self) is AsyncInterface
        return what

    @dbus_method()
    async def echo_multiple(
        self, what1: DBusStr, what2: DBusStr
    ) -> Annotated[tuple[str, str], DBusSignature("ss")]:
        assert type(self) is AsyncInterface
        return what1, what2

    @dbus_method()
    async def echo_containers(
        self,
        array: Annotated[list[str], DBusSignature("as")],
        variant: DBusVariant,
        dict_entries: DBusDict,
        struct: Annotated[
            tuple[str, tuple[str, tuple[Variant]]], DBusSignature("(s(s(v)))")
        ],
    ) -> Annotated[
        tuple[
            list[str],
            Variant,
            dict[str, Variant],
            tuple[str, tuple[str, tuple[Variant]]],
        ],
        DBusSignature("asva{sv}(s(s(v)))"),
    ]:
        assert type(self) is AsyncInterface
        return array, variant, dict_entries, struct

    @dbus_method()
    async def ping(self):
        assert type(self) is AsyncInterface

    @dbus_method(name="renamed")
    async def original_name(self):
        assert type(self) is AsyncInterface

    @dbus_method(disabled=True)
    async def not_here(self):
        assert type(self) is AsyncInterface

    @dbus_method()
    async def throws_unexpected_error(self):
        assert type(self) is AsyncInterface
        raise Exception("oops")

    @dbus_method()
    def throws_dbus_error(self):
        assert type(self) is AsyncInterface
        raise DBusError("test.error", "an error occurred")


@pytest.mark.parametrize("interface_class", [ExampleInterface, AsyncInterface])
@pytest.mark.asyncio
async def test_methods(interface_class):
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    interface = interface_class("test.interface")
    export_path = "/test/path"

    async def call(
        member, signature="", body=[], flags=MessageFlag.NONE, interface=interface.name
    ):
        msg = Message(
            destination=bus1.unique_name,
            path=export_path,
            interface=interface,
            member=member,
            signature=signature,
            body=body,
            flags=flags,
        )

        if flags & MessageFlag.NO_REPLY_EXPECTED:
            await bus2.send(msg)
            return None

        return await bus2.call(msg)

    bus1.export(export_path, interface)

    body = ["hello world"]
    reply = await call("echo", "s", body)

    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == "s"
    assert reply.body == body

    body = ["hello", "world"]
    reply = await call("echo_multiple", "ss", body)
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == "ss"
    assert reply.body == body

    body = [
        ["hello", "world"],
        Variant("v", Variant("(ss)", ("hello", "world"))),
        {"foo": Variant("t", 100)},
        ("one", ("two", (Variant("s", "three"),))),
    ]
    signature = "asva{sv}(s(s(v)))"
    SignatureTree(signature).verify(body)
    reply = await call("echo_containers", signature, body)
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == signature
    assert reply.body == body

    # Wrong interface should be a failure
    reply = await call(
        "echo_containers", signature, body, interface="org.abc.xyz.Props"
    )
    assert reply.message_type == MessageType.ERROR, reply.body[0]
    assert reply.error_name == "org.freedesktop.DBus.Error.UnknownMethod", reply.body[0]
    assert reply.body == [
        'org.abc.xyz.Props.echo_containers with signature "asva{sv}(s(s(v)))" could not be found'
    ]

    # No interface should result in finding anything that matches the member name
    # and the signature
    reply = await call("echo_containers", signature, body, interface=None)
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == signature
    assert reply.body == body

    # No interface should result in finding anything that matches the member name
    # and the signature, but in this case it will be nothing because
    # the signature is wrong
    reply = await call("echo_containers", "as", body, interface=None)
    assert reply.message_type == MessageType.ERROR, reply.body[0]
    assert reply.error_name == "org.freedesktop.DBus.Error.UnknownMethod", reply.body[0]
    assert reply.body == ['None.echo_containers with signature "as" could not be found']

    reply = await call("ping")
    assert reply.message_type == MessageType.METHOD_RETURN, reply.body[0]
    assert reply.signature == ""
    assert reply.body == []

    reply = await call("throws_unexpected_error")
    assert reply.message_type == MessageType.ERROR, reply.body[0]
    assert reply.error_name == ErrorType.SERVICE_ERROR.value, reply.body[0]
    # The body must not leak a traceback (paths, line numbers, locals) or
    # the exception's str(); only the exception class name is disclosed.
    body = reply.body[0]
    assert "Traceback" not in body, body
    assert 'File "' not in body, body
    assert "oops" not in body, body
    assert "Exception" in body, body

    reply = await call("throws_dbus_error")
    assert reply.message_type == MessageType.ERROR, reply.body[0]
    assert reply.error_name == "test.error", reply.body[0]
    assert reply.body == ["an error occurred"]

    reply = await call("ping", flags=MessageFlag.NO_REPLY_EXPECTED)
    assert reply is None

    reply = await call("throws_unexpected_error", flags=MessageFlag.NO_REPLY_EXPECTED)
    assert reply is None

    reply = await call("throws_dbus_error", flags=MessageFlag.NO_REPLY_EXPECTED)
    assert reply is None

    reply = await call("does_not_exist")
    assert reply.message_type == MessageType.ERROR, reply.body[0]
    assert reply.error_name == "org.freedesktop.DBus.Error.UnknownMethod", reply.body[0]
    assert reply.body == [
        'test.interface.does_not_exist with signature "" could not be found'
    ]

    bus1.disconnect()
    bus2.disconnect()
    await asyncio.wait_for(bus1.wait_for_disconnect(), timeout=1)
    await asyncio.wait_for(bus2.wait_for_disconnect(), timeout=1)



import io
import timeit

from dbus_fast._private.unmarshaller import Unmarshaller

#  cythonize -X language_level=3 -a -i  src/dbus_fast/_private/unmarshaller.py


bluez_mfr_data_message = (
    b"l\4\1\1x\0\0\0\232\312\n\0\225\0\0\0\1\1o\0%\0\0\0/org/bluez/hci0/dev_D0_C2_4E_08_AB_57\0\0\0\2\1s"
    b"\0\37\0\0\0org.freedesktop.DBus.Properties\0\3\1s\0\21\0\0\0PropertiesChanged\0\0\0\0\0\0\0\10\1g\0"
    b"\10sa{sv}as\0\0\0\7\1s\0\4\0\0\0:1.4\0\0\0\0\21\0\0\0org.bluez.Device1\0\0\0T\0\0\0\0\0\0\0\4\0\0\0"
    b"RSSI\0\1n\0\252\377\0\0\20\0\0\0ManufacturerData\0\5a{qv}\0$\0\0\0u\0\2ay\0\0\0\30\0\0\0B\4\1\1p\320"
    b"\302N\10\253W\322\302N\10\253V\1\0\0\0\0\0\0\0\0\0\0l\4\1\0014\0\0\0\233\312\n\0\225\0\0\0\1\1o\0%\0"
)

stream = io.BytesIO(bluez_mfr_data_message)

unmarshaller = Unmarshaller(stream)


def unmarshall_mfr_data_message():
    stream.seek(0)
    unmarshaller.unmarshall()


count = 3000000
time = timeit.Timer(unmarshall_mfr_data_message).timeit(count)
print(f"Unmarshalling {count} bluetooth ManufacturerData messages took {time} seconds")



from __future__ import annotations

import logging
from types import TracebackType
from typing import TYPE_CHECKING

from .constants import ErrorType
from .errors import DBusError
from .message import Message

if TYPE_CHECKING:
    from .message_bus import BaseMessageBus

_LOGGER = logging.getLogger(__name__)


class SendReply:
    """A context manager to send a reply to a message."""

    __slots__ = ("_bus", "_msg")

    def __init__(self, bus: BaseMessageBus, msg: Message) -> None:
        """Create a new reply context manager."""
        self._bus = bus
        self._msg = msg

    def __enter__(self) -> SendReply:
        return self

    def __call__(self, reply: Message) -> None:
        self._bus.send(reply)

    def _exit(
        self,
        exc_type: type[Exception] | None,
        exc_value: Exception | None,
        tb: TracebackType | None,
    ) -> bool:
        if exc_value:
            if isinstance(exc_value, DBusError):
                self(exc_value._as_message(self._msg))
            else:
                # Log the traceback for the operator; never send it back to
                # the caller — it discloses install paths, line numbers,
                # locals and version fingerprints to any peer that can
                # invoke a method on this service.
                _LOGGER.exception(
                    "Service interface raised an exception",
                    exc_info=(exc_type, exc_value, tb),
                )
                self(
                    Message.new_error(
                        self._msg,
                        ErrorType.SERVICE_ERROR,
                        f"The service interface raised an error: {type(exc_value).__name__}",
                    )
                )
            return True

        return False

    def __exit__(
        self,
        exc_type: type[Exception] | None,
        exc_value: Exception | None,
        tb: TracebackType | None,
    ) -> bool:
        return self._exit(exc_type, exc_value, tb)

    def send_error(self, exc: Exception) -> None:
        self._exit(exc.__class__, exc, exc.__traceback__)



import pytest

from dbus_fast import SignatureBodyMismatchError, SignatureTree, Variant
from dbus_fast._private.unmarshaller import is_compiled
from dbus_fast._private.util import signature_contains_type
from dbus_fast.errors import InternalError, InvalidSignatureError
from dbus_fast.signature import SignatureType, get_signature_tree


def assert_simple_type(signature, type_):
    assert type_.token == signature
    assert type_.signature == signature
    assert len(type_.children) == 0


def test_simple():
    tree = SignatureTree("s")
    assert len(tree.types) == 1
    assert_simple_type("s", tree.types[0])


def test_multiple_simple():
    tree = SignatureTree("sss")
    assert len(tree.types) == 3
    for i in range(3):
        assert_simple_type("s", tree.types[i])


def test_children():
    tree = SignatureTree("ss")
    assert len(tree.types) == 2
    parent = tree.types[0]
    assert parent.token == "s"
    assert parent.signature == "s"
    assert len(parent.children) == 0


def test_array():
    tree = SignatureTree("as")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.signature == "as"
    assert child.token == "a"
    assert len(child.children) == 1
    assert_simple_type("s", child.children[0])
    assert child.children[0].token == "s"
    assert child._child_0.signature == "s"
    assert child._child_1 is None
    with pytest.raises(AttributeError):
        _ = child._child_1.signature


def test_array_multiple():
    tree = SignatureTree("asasass")
    assert len(tree.types) == 4
    assert_simple_type("s", tree.types[3])
    for i in range(3):
        array_child = tree.types[i]
        assert array_child.token == "a"
        assert array_child.signature == "as"
        assert len(array_child.children) == 1
        assert_simple_type("s", array_child.children[0])


def test_array_nested():
    tree = SignatureTree("aas")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.token == "a"
    assert child.signature == "aas"
    assert len(child.children) == 1
    nested_child = child.children[0]
    assert nested_child.token == "a"
    assert nested_child.signature == "as"
    assert len(nested_child.children) == 1
    assert_simple_type("s", nested_child.children[0])


def test_simple_struct():
    tree = SignatureTree("(sss)")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.signature == "(sss)"
    assert len(child.children) == 3
    for i in range(3):
        assert_simple_type("s", child.children[i])
    assert child._child_0.signature == "s"
    assert child._child_1.signature == "s"


def test_nested_struct():
    tree = SignatureTree("(s(s(s)))")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.signature == "(s(s(s)))"
    assert child.token == "("
    assert len(child.children) == 2
    assert_simple_type("s", child.children[0])
    first_nested = child.children[1]
    assert first_nested.token == "("
    assert first_nested.signature == "(s(s))"
    assert len(first_nested.children) == 2
    assert_simple_type("s", first_nested.children[0])
    second_nested = first_nested.children[1]
    assert second_nested.token == "("
    assert second_nested.signature == "(s)"
    assert len(second_nested.children) == 1
    assert_simple_type("s", second_nested.children[0])


def test_struct_multiple():
    tree = SignatureTree("(s)(s)(s)")
    assert len(tree.types) == 3
    for i in range(3):
        child = tree.types[0]
        assert child.token == "("
        assert child.signature == "(s)"
        assert len(child.children) == 1
        assert_simple_type("s", child.children[0])


def test_array_of_structs():
    tree = SignatureTree("a(ss)")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.token == "a"
    assert child.signature == "a(ss)"
    assert len(child.children) == 1
    struct_child = child.children[0]
    assert struct_child.token == "("
    assert struct_child.signature == "(ss)"
    assert len(struct_child.children) == 2
    for i in range(2):
        assert_simple_type("s", struct_child.children[i])


def test_dict_simple():
    tree = SignatureTree("a{ss}")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.signature == "a{ss}"
    assert child.token == "a"
    assert len(child.children) == 1
    dict_child = child.children[0]
    assert dict_child.token == "{"
    assert dict_child.signature == "{ss}"
    assert len(dict_child.children) == 2
    assert_simple_type("s", dict_child.children[0])
    assert_simple_type("s", dict_child.children[1])


def test_dict_of_structs():
    tree = SignatureTree("a{s(ss)}")
    assert len(tree.types) == 1
    child = tree.types[0]
    assert child.token == "a"
    assert child.signature == "a{s(ss)}"
    assert len(child.children) == 1
    dict_child = child.children[0]
    assert dict_child.token == "{"
    assert dict_child.signature == "{s(ss)}"
    assert len(dict_child.children) == 2
    assert_simple_type("s", dict_child.children[0])
    struct_child = dict_child.children[1]
    assert struct_child.token == "("
    assert struct_child.signature == "(ss)"
    assert len(struct_child.children) == 2
    for i in range(2):
        assert_simple_type("s", struct_child.children[i])


def test_contains_type():
    tree = SignatureTree("h")
    assert signature_contains_type(tree, [0], "h")
    assert not signature_contains_type(tree, [0], "u")

    tree = SignatureTree("ah")
    assert signature_contains_type(tree, [[0]], "h")
    assert signature_contains_type(tree, [[0]], "a")
    assert not signature_contains_type(tree, [[0]], "u")

    tree = SignatureTree("av")
    body = [
        [
            Variant("u", 0),
            Variant("i", 0),
            Variant("x", 0),
            Variant("v", Variant("s", "hi")),
        ]
    ]
    assert signature_contains_type(tree, body, "u")
    assert signature_contains_type(tree, body, "x")
    assert signature_contains_type(tree, body, "v")
    assert signature_contains_type(tree, body, "s")
    assert not signature_contains_type(tree, body, "o")

    tree = SignatureTree("a{sv}")
    body = {
        "foo": Variant("h", 0),
        "bar": Variant("i", 0),
        "bat": Variant("x", 0),
        "baz": Variant("v", Variant("o", "/hi")),
    }
    for expected in "hixvso":
        assert signature_contains_type(tree, [body], expected)
    assert not signature_contains_type(tree, [body], "b")


def test_invalid_variants():
    tree = SignatureTree("a{sa{sv}}")
    s_con = {
        "type": "802-11-wireless",
        "uuid": "1234",
        "id": "SSID",
    }

    s_wifi = {
        "ssid": "SSID",
        "mode": "infrastructure",
        "hidden": True,
    }

    s_wsec = {
        "key-mgmt": "wpa-psk",
        "auth-alg": "open",
        "psk": "PASSWORD",
    }

    s_ip4 = {"method": "auto"}
    s_ip6 = {"method": "auto"}

    con = {
        "connection": s_con,
        "802-11-wireless": s_wifi,
        "802-11-wireless-security": s_wsec,
        "ipv4": s_ip4,
        "ipv6": s_ip6,
    }

    with pytest.raises(SignatureBodyMismatchError):
        tree.verify([con])


def test_variant_signature_type():
    tree = SignatureTree("as")
    var = Variant(tree.types[0], ["foo", "bar"])
    assert var.type == tree.types[0]
    assert var.value == ["foo", "bar"]
    assert var.signature == "as"

    with pytest.raises(SignatureBodyMismatchError):
        Variant(tree.types[0], "wrong")


def test_struct_accepts_tuples_or_lists():
    tree = SignatureTree("(s)")
    tree.verify([("ok",)])
    tree.verify([["ok"]])


# --- Parse-error paths in SignatureType._parse_next / SignatureTree ---


def test_parse_empty_signature_inside_container_raises():
    with pytest.raises(InvalidSignatureError, match="empty signature"):
        SignatureTree("a")


def test_parse_unknown_token_raises():
    with pytest.raises(InvalidSignatureError, match="unexpected token"):
        SignatureTree("z")


def test_parse_array_with_unterminated_struct_child_raises():
    with pytest.raises(InvalidSignatureError):
        SignatureTree("a(")


def test_parse_struct_missing_close_raises():
    with pytest.raises(InvalidSignatureError, match='missing closing "\\)"'):
        SignatureTree("(ss")


def test_parse_dict_entry_key_not_simple_raises():
    # Key must be a basic type, not a container
    with pytest.raises(InvalidSignatureError, match="simple type for dict entry key"):
        SignatureTree("a{(s)s}")


def test_parse_dict_entry_missing_close_raises():
    with pytest.raises(InvalidSignatureError, match='missing closing "}"'):
        SignatureTree("a{ss")


def test_signature_too_long_raises():
    long_sig = "s" * 256
    with pytest.raises(InvalidSignatureError, match="less than 256 characters"):
        SignatureTree(long_sig)


# --- SignatureType.__eq__ ---


def test_signature_type_eq_with_string_falls_back_to_super():
    t = SignatureTree("s").types[0]
    # Comparison against a non-SignatureType returns NotImplemented from the
    # default object __eq__, which Python interprets as False.
    assert (t == "s") is False
    assert t != 42


def test_signature_type_eq_with_other_signature_type():
    a = SignatureTree("as").types[0]
    b = SignatureTree("as").types[0]
    c = SignatureTree("ai").types[0]
    assert a == b
    assert a != c


# --- _verify_* error paths ---


def _expect_mismatch(signature, value, match=None):
    tree = SignatureTree(signature)
    with pytest.raises(SignatureBodyMismatchError, match=match):
        tree.verify([value])


def test_verify_byte_wrong_type():
    _expect_mismatch("y", "not an int", match="BYTE")


def test_verify_byte_out_of_range_low():
    _expect_mismatch("y", -1, match="between")


def test_verify_byte_out_of_range_high():
    _expect_mismatch("y", 0x100, match="between")


def test_verify_boolean_wrong_type():
    _expect_mismatch("b", 1, match="BOOLEAN")


def test_verify_int16_wrong_type():
    _expect_mismatch("n", "x", match="INT16")


def test_verify_int16_out_of_range():
    _expect_mismatch("n", 0x8000, match="between")
    _expect_mismatch("n", -0x8001, match="between")


def test_verify_uint16_wrong_type():
    _expect_mismatch("q", "x", match="UINT16")


def test_verify_uint16_out_of_range():
    _expect_mismatch("q", -1, match="between")
    _expect_mismatch("q", 0x10000, match="between")


@pytest.mark.skipif(
    is_compiled(),
    reason="_verify_int32 has a strict int annotation; Cython enforces it before the isinstance check",
)
def test_verify_int32_wrong_type():
    _expect_mismatch("i", "x", match="INT32")


def test_verify_int32_out_of_range():
    _expect_mismatch("i", 0x80000000, match="between")
    _expect_mismatch("i", -0x80000001, match="between")


def test_verify_uint32_wrong_type():
    _expect_mismatch("u", "x", match="UINT32")


def test_verify_uint32_out_of_range():
    _expect_mismatch("u", -1, match="between")
    _expect_mismatch("u", 0x100000000, match="between")


def test_verify_int64_wrong_type():
    _expect_mismatch("x", "x", match="INT64")


def test_verify_int64_out_of_range():
    _expect_mismatch("x", 1 << 63, match="between")
    _expect_mismatch("x", -(1 << 63) - 1, match="between")


def test_verify_uint64_wrong_type():
    _expect_mismatch("t", "x", match="UINT64")


def test_verify_uint64_out_of_range():
    _expect_mismatch("t", -1, match="between")
    _expect_mismatch("t", 1 << 64, match="between")


def test_verify_double_wrong_type():
    _expect_mismatch("d", "x", match="DOUBLE")


def test_verify_double_accepts_int_and_float():
    SignatureTree("d").verify([1])
    SignatureTree("d").verify([1.5])


def test_verify_unix_fd_wraps_uint32_error():
    _expect_mismatch("h", -1, match="UNIX_FD")
    _expect_mismatch("h", "x", match="UNIX_FD")


def test_verify_unix_fd_accepts_valid_uint32():
    SignatureTree("h").verify([0])
    SignatureTree("h").verify([0xFFFFFFFF])


def test_verify_object_path_invalid():
    _expect_mismatch("o", "not a path", match="OBJECT_PATH")
    _expect_mismatch("o", "missing/leading/slash", match="OBJECT_PATH")
    _expect_mismatch("o", "", match="OBJECT_PATH")


def test_verify_object_path_wrong_type():
    _expect_mismatch("o", 5, match="OBJECT_PATH")
    _expect_mismatch("o", [], match="OBJECT_PATH")
    _expect_mismatch("o", {}, match="OBJECT_PATH")


def test_verify_object_path_accepts_valid():
    SignatureTree("o").verify(["/"])
    SignatureTree("o").verify(["/com/example/Object"])


def test_verify_string_wrong_type():
    _expect_mismatch("s", 5, match="STRING")


def test_verify_signature_wrong_type():
    _expect_mismatch("g", 5, match="SIGNATURE")


def test_verify_signature_too_long():
    _expect_mismatch("g", "s" * 256, match="less than 256 bytes")


def test_verify_array_dict_wrong_container():
    _expect_mismatch("a{ss}", ["not a dict"], match="DICT_ENTRY")


def test_verify_array_bytes_wrong_type():
    _expect_mismatch("ay", [1, 2, 3], match="BYTE")


def test_verify_array_bytes_accepts_bytes_and_bytearray():
    SignatureTree("ay").verify([b"hello"])
    SignatureTree("ay").verify([bytearray(b"hello")])


def test_verify_array_wrong_container():
    _expect_mismatch("as", "not a list", match='must be Python type "list"')


def test_verify_struct_wrong_type():
    _expect_mismatch("(ss)", "not a tuple", match="STRUCT")


def test_verify_struct_wrong_length():
    _expect_mismatch(
        "(ss)", ["only one"], match="members equal to the number of struct"
    )


def test_verify_variant_wrong_type():
    _expect_mismatch("v", "not a variant", match="VARIANT")


def test_verify_none_raises():
    with pytest.raises(SignatureBodyMismatchError, match='Python type "None"'):
        SignatureType("s").verify(None)


def test_verify_token_with_no_validator_raises():
    # SignatureType built directly with a token outside the validators dict
    # (e.g. the closing-paren marker used during parsing) hits the fallback.
    t = SignatureType(")")
    with pytest.raises(Exception, match="cannot verify type with token"):
        t.verify("anything")


# --- SignatureTree.verify error paths ---


@pytest.mark.skipif(
    is_compiled(),
    reason="SignatureTree.verify has a strict list[Any] annotation; Cython enforces it before the isinstance check",
)
def test_tree_verify_body_not_list():
    tree = SignatureTree("s")
    with pytest.raises(SignatureBodyMismatchError, match="must be a list"):
        tree.verify("not a list")


def test_tree_verify_wrong_body_length():
    tree = SignatureTree("ss")
    with pytest.raises(SignatureBodyMismatchError, match="wrong number of types"):
        tree.verify(["only one"])


def test_tree_eq():
    assert SignatureTree("as") == SignatureTree("as")
    assert SignatureTree("as") != SignatureTree("ai")
    assert (SignatureTree("s") == "s") is False


# --- Variant constructor paths ---


def test_variant_from_signature_tree():
    tree = SignatureTree("s")
    v = Variant(tree, "hi")
    assert v.signature == "s"
    assert v.type is tree.types[0]
    assert v.value == "hi"


def test_variant_from_signature_string():
    v = Variant("s", "hi")
    assert v.signature == "s"
    assert v.value == "hi"


def test_variant_invalid_signature_type_raises_typeerror():
    with pytest.raises(TypeError, match="signature must be"):
        Variant(123, "hi")


def test_variant_signature_tree_with_multiple_types_raises():
    tree = SignatureTree("ss")
    with pytest.raises(ValueError, match="single complete type"):
        Variant(tree, "hi")


@pytest.mark.skipif(
    is_compiled(), reason="Variant._factory is a cdef staticmethod when Cython-compiled"
)
def test_variant_factory_skips_verification():
    tree = SignatureTree("s")
    # _factory is the internal fast path used by the unmarshaller; it bypasses
    # verification and is the only way to build a Variant with a mismatched
    # value, so we exercise that explicitly.
    v = Variant._factory(tree, "hi")
    assert v.signature == "s"
    assert v.type is tree.root_type
    assert v.value == "hi"


def test_variant_eq():
    assert Variant("s", "hi") == Variant("s", "hi")
    assert Variant("s", "hi") != Variant("s", "bye")
    assert Variant("s", "hi") != Variant("o", "/hi")
    assert (Variant("s", "hi") == "hi") is False


def test_variant_repr():
    v = Variant("s", "hi")
    text = repr(v)
    assert "Variant" in text
    assert "'s'" in text
    assert "hi" in text


def test_variant_skip_verify_flag():
    # verify=False lets the caller construct an invalid Variant without raising;
    # used as a perf escape hatch by trusted producers.
    v = Variant("s", 123, verify=False)
    assert v.value == 123


def test_verify_raises_internal_error_when_validator_missing(monkeypatch):
    sig_type = SignatureType("i")
    monkeypatch.delitem(SignatureType.validators, "i")
    with pytest.raises(InternalError):
        sig_type.verify(123)


def test_get_signature_tree_cache_is_bounded() -> None:
    """The lru_cache on get_signature_tree must declare a finite maxsize."""
    info = get_signature_tree.cache_info()
    assert info.maxsize is not None
    assert info.maxsize <= 8192


def test_get_signature_tree_cache_evicts_under_unique_stream() -> None:
    """Streaming more unique signatures than maxsize must not grow the cache."""
    info = get_signature_tree.cache_info()
    maxsize = info.maxsize
    assert maxsize is not None

    # Encode the index in 14 bits across two DBus type chars (y, i) so each
    # signature is unique, short, and well under the 255-char signature
    # length cap.
    overflow = maxsize * 2
    width = overflow.bit_length()
    for n in range(overflow):
        sig = f"{n:b}".zfill(width).translate(str.maketrans("01", "yi"))
        get_signature_tree(sig)

    assert get_signature_tree.cache_info().currsize <= maxsize



import io
import timeit

from dbus_fast._private.unmarshaller import Unmarshaller

#  cythonize -X language_level=3 -a -i  src/dbus_fast/_private/unmarshaller.py


bluez_interfaces_removed_message = (
    b'l\4\1\1\222\0\0\0\377@-\0~\0\0\0\1\1o\0\1\0\0\0/\0\0\0\0\0\0\0\2\1s\0"\0\0\0'
    b"org.freedesktop.DBus.ObjectManager\0\0\0\0\0\0\3\1s\0\21\0\0\0InterfacesRemoved"
    b"\0\0\0\0\0\0\0\10\1g\0\3oas\0\0\0\0\0\0\0\0\7\1s\0\5\0\0\0:1.12\0\0\0%\0\0\0"
    b"/org/bluez/hci0/dev_5F_13_47_38_26_55\0\0\0b\0\0\0\37\0\0\0org.freedesktop.DBus"
    b".Properties\0#\0\0\0org.freedesktop.DBus.Introspectable\0\21\0\0\0org.bluez.Dev"
    b"ice1\0"
)


stream = io.BytesIO(bluez_interfaces_removed_message)
unmarshaller = Unmarshaller(stream)


def unmarshall_interfaces_removed_message():
    stream.seek(0)
    unmarshaller.unmarshall()


count = 3000000
time = timeit.Timer(unmarshall_interfaces_removed_message).timeit(count)
print(f"Unmarshalling {count} bluetooth InterfacesRemoved messages took {time} seconds")



class DBusFastError(Exception):
    """Common base class for all dbus-fast exceptions.

    Catch this to handle any error raised by dbus-fast regardless of which
    Python built-in exception type the specific error also derives from.
    Existing ``except ValueError`` / ``except TypeError`` handlers continue
    to work because individual error classes still inherit from those.
    """


class SignatureBodyMismatchError(ValueError, DBusFastError):
    pass


class InvalidSignatureError(ValueError, DBusFastError):
    pass


class InvalidAddressError(ValueError, DBusFastError):
    pass


class AuthError(DBusFastError):
    pass


class InternalError(RuntimeError, DBusFastError):
    """Indicates a bug inside dbus-fast itself."""


class InvalidMessageError(ValueError, DBusFastError):
    pass


class InvalidIntrospectionError(ValueError, DBusFastError):
    pass


class InterfaceNotFoundError(DBusFastError):
    pass


class SignalDisabledError(DBusFastError):
    pass


class InvalidBusNameError(TypeError, DBusFastError):
    def __init__(self, name: str) -> None:
        super().__init__(f"invalid bus name: {name}")


class InvalidObjectPathError(TypeError, DBusFastError):
    def __init__(self, path: str) -> None:
        super().__init__(f"invalid object path: {path}")


class InvalidInterfaceNameError(TypeError, DBusFastError):
    def __init__(self, name: str) -> None:
        super().__init__(f"invalid interface name: {name}")


class InvalidMemberNameError(TypeError, DBusFastError):
    def __init__(self, member: str) -> None:
        super().__init__(f"invalid member name: {member}")


from .constants import ErrorType, MessageType  # noqa: E402
from .message import Message  # noqa: E402
from .validators import assert_interface_name_valid  # noqa: E402


class DBusError(DBusFastError):
    def __init__(
        self, type_: ErrorType | str, text: str, reply: Message | None = None
    ) -> None:
        super().__init__(text)

        if type(type_) is ErrorType:
            type_ = type_.value

        assert_interface_name_valid(type_)  # type: ignore[arg-type]
        if reply is not None and type(reply) is not Message:
            raise TypeError("reply must be of type Message")

        self.type = type_
        self.text = text
        self.reply = reply

    @staticmethod
    def _from_message(msg: Message) -> "DBusError":
        assert msg.message_type == MessageType.ERROR
        return DBusError(msg.error_name or "unknown", msg.body[0], reply=msg)

    def _as_message(self, msg: Message) -> Message:
        return Message.new_error(msg, self.type, self.text)



# cython: freethreading_compatible = True

from typing import Any

from .signature import Variant


def unpack_variants(data: Any) -> Any:
    """Unpack variants and remove signature info.

    This function should only be used to unpack
    unmarshalled data as the checks are not
    idiomatic.
    """
    return _unpack_variants(data)


def _unpack_variants(data: Any) -> Any:
    if type(data) is dict:
        return {k: _unpack_variants(v) for k, v in data.items()}
    if type(data) is list:
        return [_unpack_variants(item) for item in data]
    if type(data) is tuple:
        return tuple(_unpack_variants(item) for item in data)
    if type(data) is Variant:
        var = data
        return _unpack_variants(var.value)
    return data



import asyncio

import pytest

from dbus_fast import Message, MessageType
from dbus_fast import introspection as intr
from dbus_fast.aio import MessageBus
from dbus_fast.service import ServiceInterface, dbus_method

standard_interfaces_count = len(intr.Node.default().interfaces)


class ExampleInterface(ServiceInterface):
    def __init__(self, name):
        self._method_called = False
        super().__init__(name)

    @dbus_method()
    def some_method(self):
        self._method_called = True


@pytest.mark.asyncio
async def test_export_unexport():
    interface = ExampleInterface("test.interface")
    interface2 = ExampleInterface("test.interface2")

    export_path = "/test/path"
    export_path2 = "/test/path/child"

    bus = await MessageBus().connect()
    bus.export(export_path, interface)

    with pytest.raises(ValueError):
        # Already exported
        bus.export(export_path, interface)

    assert export_path in bus._path_exports
    assert len(bus._path_exports[export_path]) == 1
    assert bus._path_exports[export_path][interface.name] is interface
    assert len(ServiceInterface._get_buses(interface)) == 1

    bus.export(export_path2, interface2)

    node = bus._introspect_export_path(export_path)
    assert len(node.interfaces) == standard_interfaces_count + 1
    assert len(node.nodes) == 1
    # relative path
    assert node.nodes[0].name == "child"

    bus.unexport(export_path, interface)
    assert export_path not in bus._path_exports
    assert len(ServiceInterface._get_buses(interface)) == 0

    bus.export(export_path2, interface)
    assert len(bus._path_exports[export_path2]) == 2

    # test unexporting the whole path
    bus.unexport(export_path2)
    assert not bus._path_exports
    assert not ServiceInterface._get_buses(interface)
    assert not ServiceInterface._get_buses(interface2)

    # test unexporting by name
    bus.export(export_path, interface)
    bus.unexport(export_path, interface.name)
    assert not bus._path_exports
    assert not ServiceInterface._get_buses(interface)

    # test unexporting by ServiceInterface
    bus.export(export_path, interface)
    bus.unexport(export_path, interface)
    assert not bus._path_exports
    assert not ServiceInterface._get_buses(interface)

    with pytest.raises(TypeError):
        bus.unexport(export_path, object())

    node = bus._introspect_export_path("/path/doesnt/exist")
    assert type(node) is intr.Node
    assert not node.interfaces
    assert not node.nodes

    # Should to nothing
    bus.unexport("/path/doesnt/exist", interface)

    bus.disconnect()
    await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_export_twice_raises():
    bus = await MessageBus().connect()

    interface = ExampleInterface("test.interface")

    export_path = "/test/path"
    export_path2 = "/test/path/child"

    bus.export(export_path, interface)

    with pytest.raises(
        ValueError, match="instance cannot be added to the same bus twice"
    ):
        bus.export(export_path2, interface)

    bus.disconnect()
    await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_export_alias():
    bus = await MessageBus().connect()

    interface = ExampleInterface("test.interface")
    interface2 = ExampleInterface("test.interface")

    export_path = "/test/path"
    export_path2 = "/test/path/child"

    bus.export(export_path, interface)
    bus.export(export_path2, interface2)

    result = await bus.call(
        Message(
            destination=bus.unique_name,
            path=export_path,
            interface="test.interface",
            member="some_method",
        )
    )
    assert result.message_type is MessageType.METHOD_RETURN, result.body[0]

    assert interface._method_called
    interface._method_called = False

    result = await bus.call(
        Message(
            destination=bus.unique_name,
            path=export_path2,
            interface="test.interface",
            member="some_method",
        )
    )
    assert result.message_type is MessageType.METHOD_RETURN, result.body[0]
    assert interface2._method_called

    bus.disconnect()
    await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_export_introspection():
    interface = ExampleInterface("test.interface")
    interface2 = ExampleInterface("test.interface2")

    export_path = "/test/path"
    export_path2 = "/test/path/child"

    bus = await MessageBus().connect()
    bus.export(export_path, interface)
    bus.export(export_path2, interface2)

    root = bus._introspect_export_path("/")
    assert len(root.nodes) == 1

    bus.disconnect()
    await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)



"""cdefs for unmarshaller.py"""

import cython

from ..message cimport Message
from ..signature cimport SignatureTree, SignatureType, Variant

cdef bint TYPE_CHECKING

cdef object MAX_UNIX_FDS_SIZE
cdef object ARRAY
cdef object UNIX_FDS_CMSG_LENGTH
cdef object SOL_SOCKET
cdef object SCM_RIGHTS
cdef object MESSAGE_FLAG_INTENUM

cdef unsigned int UINT32_SIZE
cdef unsigned int INT16_SIZE
cdef unsigned int UINT16_SIZE

cdef unsigned int HEADER_ARRAY_OF_STRUCT_SIGNATURE_POSITION
cdef unsigned int HEADER_SIGNATURE_SIZE
cdef unsigned int LITTLE_ENDIAN
cdef unsigned int BIG_ENDIAN
cdef unsigned int PROTOCOL_VERSION
cdef unsigned int _MAX_MESSAGE_SIZE


cdef unsigned int HEADER_PATH_IDX
cdef unsigned int HEADER_INTERFACE_IDX
cdef unsigned int HEADER_MEMBER_IDX
cdef unsigned int HEADER_ERROR_NAME_IDX
cdef unsigned int HEADER_REPLY_SERIAL_IDX
cdef unsigned int HEADER_DESTINATION_IDX
cdef unsigned int HEADER_SENDER_IDX
cdef unsigned int HEADER_SIGNATURE_IDX
cdef unsigned int HEADER_UNIX_FDS_IDX

cdef cython.list HEADER_IDX_TO_ARG_NAME

cdef str UINT32_CAST
cdef str INT16_CAST
cdef str UINT16_CAST

cdef bint SYS_IS_LITTLE_ENDIAN
cdef bint SYS_IS_BIG_ENDIAN

cdef object UNPACK_HEADER_LITTLE_ENDIAN
cdef object UNPACK_HEADER_BIG_ENDIAN

cdef object UINT32_UNPACK_LITTLE_ENDIAN
cdef object UINT32_UNPACK_BIG_ENDIAN

cdef object INT16_UNPACK_LITTLE_ENDIAN
cdef object INT16_UNPACK_BIG_ENDIAN

cdef object UINT16_UNPACK_LITTLE_ENDIAN
cdef object UINT16_UNPACK_BIG_ENDIAN

cdef cython.dict MESSAGE_TYPE_MAP
cdef cython.dict MESSAGE_FLAG_MAP
cdef dict HEADER_MESSAGE_ARG_NAME


cdef SignatureTree SIGNATURE_TREE_EMPTY
cdef SignatureTree SIGNATURE_TREE_B
cdef SignatureTree SIGNATURE_TREE_N
cdef SignatureTree SIGNATURE_TREE_O
cdef SignatureTree SIGNATURE_TREE_S
cdef SignatureTree SIGNATURE_TREE_U
cdef SignatureTree SIGNATURE_TREE_Y

cdef SignatureTree SIGNATURE_TREE_AS
cdef SignatureType SIGNATURE_TREE_AS_TYPES_0
cdef SignatureTree SIGNATURE_TREE_AO
cdef SignatureType SIGNATURE_TREE_AO_TYPES_0
cdef SignatureTree SIGNATURE_TREE_A_SV
cdef SignatureType SIGNATURE_TREE_A_SV_TYPES_0
cdef SignatureTree SIGNATURE_TREE_SA_SV_AS
cdef SignatureType SIGNATURE_TREE_SA_SV_AS_TYPES_1
cdef SignatureType SIGNATURE_TREE_SA_SV_AS_TYPES_2
cdef SignatureTree SIGNATURE_TREE_OAS
cdef SignatureType SIGNATURE_TREE_OAS_TYPES_1
cdef SignatureTree SIGNATURE_TREE_OA_SA_SV
cdef SignatureType SIGNATURE_TREE_OA_SA_SV_TYPES_1
cdef SignatureTree SIGNATURE_TREE_AY
cdef SignatureType SIGNATURE_TREE_AY_TYPES_0
cdef SignatureTree SIGNATURE_TREE_A_QV
cdef SignatureType SIGNATURE_TREE_A_QV_TYPES_0
cdef SignatureTree SIGNATURE_TREE_A_OA_SA_SV
cdef SignatureType SIGNATURE_TREE_A_OA_SA_SV_TYPES_0


cdef unsigned int TOKEN_B_AS_INT
cdef unsigned int TOKEN_U_AS_INT
cdef unsigned int TOKEN_Y_AS_INT
cdef unsigned int TOKEN_A_AS_INT
cdef unsigned int TOKEN_O_AS_INT
cdef unsigned int TOKEN_S_AS_INT
cdef unsigned int TOKEN_G_AS_INT
cdef unsigned int TOKEN_N_AS_INT
cdef unsigned int TOKEN_X_AS_INT
cdef unsigned int TOKEN_T_AS_INT
cdef unsigned int TOKEN_D_AS_INT
cdef unsigned int TOKEN_Q_AS_INT
cdef unsigned int TOKEN_V_AS_INT
cdef unsigned int TOKEN_LEFT_CURLY_AS_INT
cdef unsigned int TOKEN_LEFT_PAREN_AS_INT

cdef object MARSHALL_STREAM_END_ERROR
cdef object DEFAULT_BUFFER_SIZE

cdef Variant VARIANT_BOOL_TRUE
cdef Variant VARIANT_BOOL_FALSE

cdef list _EMPTY_HEADERS

cdef cython.uint EAGAIN
cdef cython.uint EWOULDBLOCK

cdef get_signature_tree


cdef unsigned int _ustr_uint32(const unsigned char * buf, unsigned int offset, unsigned int endian) noexcept

cdef short _ustr_int16(const unsigned char * buf, unsigned int offset, unsigned int endian) noexcept

cdef unsigned short _ustr_uint16(const unsigned char * buf, unsigned int offset, unsigned int endian) noexcept


cdef class Unmarshaller:

    cdef list _unix_fds
    cdef bytearray _buf
    cdef Py_ssize_t _buf_len
    cdef const unsigned char * _buf_ustr
    cdef unsigned int _pos
    cdef object _stream
    cdef object _sock
    cdef Message _message
    cdef object _readers
    cdef unsigned int _body_len
    cdef unsigned int _serial
    cdef unsigned int _header_len
    cdef object _message_type
    cdef object _flag
    cdef unsigned int _msg_len
    cdef object _uint32_unpack
    cdef object _int16_unpack
    cdef object _uint16_unpack
    cdef object _stream_reader
    cdef object _sock_with_fds_reader
    cdef object _sock_without_fds_reader
    cdef bint _negotiate_unix_fd
    cdef bint _read_complete
    cdef unsigned int _endian

    @cython.locals(to_clear=Py_ssize_t)
    cdef _next_message(self)

    cdef bint _has_another_message_in_buffer(self)

    @cython.locals(
        msg=cython.bytes,
        recv=cython.tuple,
        errno=cython.uint
    )
    cdef void _read_sock_with_fds(self, unsigned int pos, unsigned int missing_bytes) except *

    @cython.locals(
        data=cython.bytes,
        errno=cython.uint
    )
    cdef void _read_sock_without_fds(self, unsigned int pos) except *

    @cython.locals(
        data=cython.bytes
    )
    cdef void _read_stream(self, unsigned int pos, unsigned int missing_bytes) except *

    cdef void _read_to_pos(self, unsigned int pos) except *

    cpdef read_boolean(self, SignatureType type_)

    cdef bint _read_boolean(self)

    cpdef read_uint32_unpack(self, SignatureType type_)

    cdef unsigned int _read_uint32_unpack(self)

    cpdef read_int16_unpack(self, SignatureType type_)

    cdef int _read_int16_unpack(self)

    cpdef read_uint16_unpack(self, SignatureType type_)

    cdef unsigned int _read_uint16_unpack(self)

    cpdef read_string_unpack(self, SignatureType type_)

    @cython.locals(
        str_start=cython.uint,
    )
    cdef str _read_string_unpack(self)

    @cython.locals(
        tree=SignatureTree,
        token_as_int=cython.uint,
        signature_len=cython.uint,
        o=cython.ulong,
        var=Variant,
    )
    cdef Variant _read_variant(self)

    @cython.locals(
        beginning_pos=cython.ulong,
        array_length=cython.uint,
        children=cython.list,
        child_type=SignatureType,
        child_0=SignatureType,
        child_1=SignatureType,
        token_as_int=cython.uint,
    )
    cpdef object read_array(self, SignatureType type_)

    cpdef read_signature(self, SignatureType type_)

    @cython.locals(
        o=cython.ulong,
        signature_len=cython.uint,
    )
    cdef str _read_signature(self)

    @cython.locals(
        endian=cython.uint,
        buffer=cython.bytearray,
        protocol_version=cython.uint,
        key=cython.str,
        ustring="const unsigned char *",
    )
    cdef void _read_header(self) except *

    @cython.locals(
        body=list,
        header_fields=list,
        token_as_int="unsigned int",
        signature=str,
        tree=SignatureTree,
        message=Message,
        reply_serial=object,
    )
    cdef void _read_body(self) except *

    cdef Message _unmarshall(self)

    cpdef unmarshall(self)

    @cython.locals(
        beginning_pos=cython.ulong,
        o=cython.ulong,
        token_as_int=cython.uint,
        signature_len=cython.uint,
        headers=cython.list
    )
    cdef cython.list _header_fields(self, unsigned int header_length)



import os
from collections.abc import Generator
from unittest.mock import patch

import pytest

from dbus_fast.constants import ErrorType, MessageType
from dbus_fast.errors import DBusError
from dbus_fast.message import Message
from dbus_fast.message_bus import BaseMessageBus
from dbus_fast.send_reply import SendReply


@pytest.fixture(autouse=True)
def mock_address() -> Generator[None, None, None]:
    original_address = os.environ.get("DBUS_SESSION_BUS_ADDRESS")
    os.environ["DBUS_SESSION_BUS_ADDRESS"] = "unix:path=/dev/null"
    yield
    if original_address is None:
        del os.environ["DBUS_SESSION_BUS_ADDRESS"]
    else:
        os.environ["DBUS_SESSION_BUS_ADDRESS"] = original_address


@pytest.fixture
def send_reply_setup() -> Generator[
    tuple[BaseMessageBus, Message, list[Message]], None, None
]:
    messages: list[Message] = []

    class MockClosable:
        def close(self) -> None:
            pass

    class MockBus(BaseMessageBus):
        def send(self, msg: Message) -> None:
            messages.append(msg)

        def send_message(self, msg: Message) -> None:
            messages.append(msg)

        def _setup_socket(self) -> None:
            self._sock = MockClosable()  # type: ignore
            self._stream = MockClosable()  # type: ignore

    with patch("socket.socket.connect"):
        bus = MockBus()
    msg = Message(
        path="/test/path", interface="test.interface", member="test_member", serial=1
    )
    try:
        yield bus, msg, messages
    finally:
        bus.disconnect()
        bus._finalize(None)


def test_send_reply_exception(
    send_reply_setup: tuple[BaseMessageBus, Message, list[Message]],
) -> None:
    """Test that SendReply sends an error message when DBusError is raised."""
    bus, msg, messages = send_reply_setup
    send_reply = SendReply(bus, msg)

    with send_reply:
        raise DBusError(ErrorType.DISCONNECTED, "Disconnected", None)

    assert len(messages) == 1
    assert messages[0].message_type == MessageType.ERROR
    assert messages[0].error_name == "org.freedesktop.DBus.Error.Disconnected"
    assert messages[0].reply_serial == 1


def test_send_reply_generic_exception(
    send_reply_setup: tuple[BaseMessageBus, Message, list[Message]],
) -> None:
    """Non-DBusError exceptions become SERVICE_ERROR replies."""
    bus, msg, messages = send_reply_setup
    send_reply = SendReply(bus, msg)

    with send_reply:
        raise RuntimeError("boom")

    assert len(messages) == 1
    assert messages[0].message_type == MessageType.ERROR
    assert messages[0].error_name == ErrorType.SERVICE_ERROR.value
    assert messages[0].reply_serial == 1
    # The wire body must disclose only the exception class name — never the
    # str() of the exception (might contain caller data) and never a Python
    # traceback (paths, line numbers, locals).
    body = messages[0].body[0]
    assert "boom" not in body, body
    assert "Traceback" not in body, body
    assert 'File "' not in body, body
    assert "RuntimeError" in body, body


def test_send_reply_send_error(
    send_reply_setup: tuple[BaseMessageBus, Message, list[Message]],
) -> None:
    """send_error() routes an exception through the same error reply path."""
    bus, msg, messages = send_reply_setup
    send_reply = SendReply(bus, msg)

    send_reply.send_error(DBusError(ErrorType.FAILED, "explicit failure", None))

    assert len(messages) == 1
    assert messages[0].message_type == MessageType.ERROR
    assert messages[0].error_name == "org.freedesktop.DBus.Error.Failed"
    assert messages[0].reply_serial == 1


def test_send_reply_no_exception_returns_false(
    send_reply_setup: tuple[BaseMessageBus, Message, list[Message]],
) -> None:
    """Exiting with no exception must not suppress and must not send a reply."""
    bus, msg, messages = send_reply_setup
    send_reply = SendReply(bus, msg)

    assert send_reply._exit(None, None, None) is False
    assert messages == []


def test_send_reply_happy_path(
    send_reply_setup: tuple[BaseMessageBus, Message, list[Message]],
) -> None:
    """Test that SendReply sends a message."""
    bus, msg, messages = send_reply_setup
    send_reply = SendReply(bus, msg)

    with send_reply as reply:
        reply(msg)

    assert len(messages) == 1
    assert messages[0].message_type == MessageType.METHOD_CALL
    assert messages[0].error_name is None



"""cdefs for signature.py"""

import cython


cdef class SignatureType:

    cdef public str token
    cdef public unsigned int token_as_int
    cdef public list children
    cdef str _signature
    cdef public SignatureType _child_0
    cdef public SignatureType _child_1


cdef class SignatureTree:

    cdef public str signature
    cdef public list types
    cdef public SignatureType root_type


cdef class Variant:

    cdef public SignatureType type
    cdef public str signature
    cdef public object value

    @cython.locals(self=Variant)
    @staticmethod
    cdef Variant _factory(SignatureTree signature_tree, object value)



import enum
import os

from .errors import AuthError

UID_NOT_SPECIFIED = -1

# SASL auth lines are tiny in practice; cap the buffer so a hostile peer
# cannot exhaust memory by streaming bytes without CRLF. Dual-name pattern
# (see CLAUDE.md) so a future cdef alias in a .pxd stays Python-importable.
MAX_AUTH_LINE = 16 * 1024
_MAX_AUTH_LINE = MAX_AUTH_LINE

AUTH_READ_CHUNK = 1024
_AUTH_READ_CHUNK = AUTH_READ_CHUNK

# The auth interface here is unstable. I would like to eventually open this up
# for people to define their own custom authentication protocols, but I'm not
# familiar with what's needed for that exactly. To work with any message bus
# implementation would require abstracting out all the IO. Async operations
# might be challenging because different IO backends have different ways of
# doing that. I might just end up giving the raw socket and leaving it all up
# to the user, but it would be nice to have a little guidance in the interface
# since a lot of it is strongly specified. If you have a need for this, contact
# the project maintainer to help stabilize this interface.


class _AuthResponse(enum.Enum):
    OK = "OK"
    REJECTED = "REJECTED"
    DATA = "DATA"
    ERROR = "ERROR"
    AGREE_UNIX_FD = "AGREE_UNIX_FD"

    @classmethod
    def parse(cls, line: str) -> tuple["_AuthResponse", list[str]]:
        args = line.split(" ")
        response = cls(args[0])
        return response, args[1:]


# UNSTABLE
class Authenticator:
    """The base class for authenticators for :class:`MessageBus <dbus_fast.message_bus.BaseMessageBus>` authentication.

    In the future, the library may allow extending this class for custom authentication protocols.

    :seealso: https://dbus.freedesktop.org/doc/dbus-specification.html#auth-protocol
    """

    def _authentication_start(self, negotiate_unix_fd: bool = False) -> str:
        raise NotImplementedError(
            "authentication_start() must be implemented in the inheriting class"
        )

    def _receive_line(self, line: str) -> str:
        raise NotImplementedError(
            "receive_line() must be implemented in the inheriting class"
        )

    @staticmethod
    def _format_line(line: str) -> bytes:
        return f"{line}\r\n".encode()


class AuthExternal(Authenticator):
    """An authenticator class for the external auth protocol for use with the
    :class:`MessageBus <dbus_fast.message_bus.BaseMessageBus>`.

    :param uid: The uid to use when connecting to the message bus. Use UID_NOT_SPECIFIED to use the uid known to the kernel.
    :vartype uid: int

    :sealso: https://dbus.freedesktop.org/doc/dbus-specification.html#auth-protocol
    """

    def __init__(self, uid: int | None = None) -> None:
        self.negotiate_unix_fd: bool = False
        self.negotiating_fds: bool = False
        self.uid: int | None = uid

    def _authentication_start(self, negotiate_unix_fd: bool = False) -> str:
        self.negotiate_unix_fd = negotiate_unix_fd
        uid = self.uid
        if uid == UID_NOT_SPECIFIED:
            return "AUTH EXTERNAL"
        if uid is None:
            uid = os.getuid()
        hex_uid = str(uid).encode().hex()
        return f"AUTH EXTERNAL {hex_uid}"

    def _receive_line(self, line: str) -> str:
        response, args = _AuthResponse.parse(line)

        if response is _AuthResponse.OK:
            if self.negotiate_unix_fd:
                self.negotiating_fds = True
                return "NEGOTIATE_UNIX_FD"
            return "BEGIN"

        if response is _AuthResponse.AGREE_UNIX_FD:
            return "BEGIN"

        if response is _AuthResponse.DATA and self.uid == UID_NOT_SPECIFIED:
            return "DATA"

        raise AuthError(f"authentication failed: {response.value}: {args}")


class AuthAnonymous(Authenticator):
    """An authenticator class for the anonymous auth protocol for use with the
    :class:`MessageBus <dbus_fast.message_bus.BaseMessageBus>`.

    :sealso: https://dbus.freedesktop.org/doc/dbus-specification.html#auth-protocol
    """

    def _authentication_start(self, negotiate_unix_fd: bool = False) -> str:
        if negotiate_unix_fd:
            raise AuthError(
                "anonymous authentication does not support negotiating unix fds right now"
            )

        return "AUTH ANONYMOUS"

    def _receive_line(self, line: str) -> str:
        response, args = _AuthResponse.parse(line)

        if response != _AuthResponse.OK:
            raise AuthError(f"authentication failed: {response.value}: {args}")

        return "BEGIN"


# The following line provides backwards compatibility, remove at some point? --jrd
AuthAnnonymous = AuthAnonymous



import timeit

from dbus_fast import Variant, unpack_variants

# cythonize -X language_level=3 -a -i  src/dbus_fast/unpack.py


message = {
    "/org/bluez/hci0": {
        "org.bluez.Adapter1": {
            "Address": Variant("s", "00:E0:4C:2A:25:63"),
            "AddressType": Variant("s", "public"),
            "Alias": Variant("s", "BlueZ 5.63"),
            "Class": Variant("u", 2883584),
            "Discoverable": Variant("b", False),
            "DiscoverableTimeout": Variant("u", 180),
            "Discovering": Variant("b", True),
            "Modalias": Variant("s", "usb:v1D6Bp0246d053F"),
            "Name": Variant("s", "BlueZ 5.63"),
            "Pairable": Variant("b", False),
            "PairableTimeout": Variant("u", 0),
            "Powered": Variant("b", True),
            "Roles": Variant("as", ["central", "peripheral"]),
            "UUIDs": Variant(
                "as",
                [
                    "0000110e-0000-1000-8000-00805f9b34fb",
                    "0000110a-0000-1000-8000-00805f9b34fb",
                    "00001200-0000-1000-8000-00805f9b34fb",
                    "0000110b-0000-1000-8000-00805f9b34fb",
                    "00001108-0000-1000-8000-00805f9b34fb",
                    "0000110c-0000-1000-8000-00805f9b34fb",
                    "00001800-0000-1000-8000-00805f9b34fb",
                    "00001801-0000-1000-8000-00805f9b34fb",
                    "0000180a-0000-1000-8000-00805f9b34fb",
                    "00001112-0000-1000-8000-00805f9b34fb",
                ],
            ),
        },
        "org.bluez.GattManager1": {},
        "org.bluez.LEAdvertisingManager1": {
            "ActiveInstances": Variant("y", 0),
            "SupportedIncludes": Variant(
                "as", ["tx-power", "appearance", "local-name"]
            ),
            "SupportedInstances": Variant("y", 4),
            "SupportedSecondaryChannels": Variant("as", ["1M", "2M", "Coded"]),
        },
        "org.bluez.Media1": {},
        "org.bluez.NetworkServer1": {},
        "org.freedesktop.DBus.Introspectable": {},
        "org.freedesktop.DBus.Properties": {},
    },
}


def unpack_managed_objects():
    unpack_variants(message)


count = 3000000
time = timeit.Timer(unpack_managed_objects).timeit(count)
print(f"Unpacked {count} get managed objects messages took {time} seconds")



import asyncio
from typing import Annotated

import pytest

from dbus_fast import Message, MessageType
from dbus_fast.aio import MessageBus
from dbus_fast.annotations import DBusInt32, DBusSignature, DBusStr
from dbus_fast.constants import PropertyAccess
from dbus_fast.service import (
    ServiceInterface,
    SignalDisabledError,
    dbus_property,
    dbus_signal,
)
from dbus_fast.signature import Variant


class ExampleInterface(ServiceInterface):
    def __init__(self, name: str) -> None:
        super().__init__(name)

    @dbus_signal()
    def signal_empty(self):
        assert type(self) is ExampleInterface

    @dbus_signal()
    def signal_simple(self) -> DBusStr:
        assert type(self) is ExampleInterface
        return "hello"

    @dbus_signal()
    def signal_multiple(self) -> Annotated[tuple[str, str], DBusSignature("ss")]:
        assert type(self) is ExampleInterface
        return "hello", "world"

    @dbus_signal(name="renamed")
    def original_name(self):
        assert type(self) is ExampleInterface

    @dbus_signal(disabled=True)
    def signal_disabled(self):
        assert type(self) is ExampleInterface

    @dbus_property(access=PropertyAccess.READ)
    def test_prop(self) -> DBusInt32:
        return 42


class SecondExampleInterface(ServiceInterface):
    def __init__(self, name: str) -> None:
        super().__init__(name)

    @dbus_property(access=PropertyAccess.READ)
    def str_prop(self) -> DBusStr:
        return "abc"

    @dbus_property(access=PropertyAccess.READ)
    def list_prop(self) -> Annotated[list[int], DBusSignature("ai")]:
        return [1, 2, 3]


class ExpectMessage:
    def __init__(self, bus1, bus2, interface_name, timeout=1):
        self.future = asyncio.get_running_loop().create_future()
        self.bus1 = bus1
        self.bus2 = bus2
        self.interface_name = interface_name
        self.timeout = timeout
        self.timeout_task = None

    def message_handler(self, msg):
        if msg.sender == self.bus1.unique_name and msg.interface == self.interface_name:
            self.timeout_task.cancel()
            self.future.set_result(msg)
            return True

    def timeout_cb(self):
        self.future.set_exception(TimeoutError)

    async def __aenter__(self):
        self.bus2.add_message_handler(self.message_handler)
        self.timeout_task = asyncio.get_running_loop().call_later(
            self.timeout, self.timeout_cb
        )

        return self.future

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.bus2.remove_message_handler(self.message_handler)


def assert_signal_ok(signal, export_path, member, signature, body):
    assert signal.message_type == MessageType.SIGNAL
    assert signal.path == export_path
    assert signal.member == member
    assert signal.signature == signature
    assert signal.body == body


@pytest.mark.asyncio
async def test_signals():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    interface = ExampleInterface("test.interface")
    export_path = "/test/path"
    bus1.export(export_path, interface)

    await bus2.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="AddMatch",
            signature="s",
            body=[f"sender={bus1.unique_name}"],
        )
    )

    async with ExpectMessage(bus1, bus2, interface.name) as expected_signal:
        interface.signal_empty()
        assert_signal_ok(
            signal=await expected_signal,
            export_path=export_path,
            member="signal_empty",
            signature="",
            body=[],
        )

    async with ExpectMessage(bus1, bus2, interface.name) as expected_signal:
        interface.original_name()
        assert_signal_ok(
            signal=await expected_signal,
            export_path=export_path,
            member="renamed",
            signature="",
            body=[],
        )

    async with ExpectMessage(bus1, bus2, interface.name) as expected_signal:
        interface.signal_simple()
        assert_signal_ok(
            signal=await expected_signal,
            export_path=export_path,
            member="signal_simple",
            signature="s",
            body=["hello"],
        )

    async with ExpectMessage(bus1, bus2, interface.name) as expected_signal:
        interface.signal_multiple()
        assert_signal_ok(
            signal=await expected_signal,
            export_path=export_path,
            member="signal_multiple",
            signature="ss",
            body=["hello", "world"],
        )

    with pytest.raises(SignalDisabledError):
        interface.signal_disabled()

    bus1.disconnect()
    bus2.disconnect()
    await asyncio.wait_for(bus1.wait_for_disconnect(), timeout=1)
    await asyncio.wait_for(bus2.wait_for_disconnect(), timeout=1)


@pytest.mark.asyncio
async def test_interface_add_remove_signal():
    bus1 = await MessageBus().connect()
    bus2 = await MessageBus().connect()

    await bus2.call(
        Message(
            destination="org.freedesktop.DBus",
            path="/org/freedesktop/DBus",
            interface="org.freedesktop.DBus",
            member="AddMatch",
            signature="s",
            body=[f"sender={bus1.unique_name}"],
        )
    )

    first_interface = ExampleInterface("test.interface.first")
    second_interface = SecondExampleInterface("test.interface.second")
    export_path = "/test/path"

    # add first interface
    async with ExpectMessage(
        bus1, bus2, "org.freedesktop.DBus.ObjectManager"
    ) as expected_signal:
        bus1.export(export_path, first_interface)
        assert_signal_ok(
            signal=await expected_signal,
            export_path=export_path,
            member="InterfacesAdded",
            signature="oa{sa{sv}}",
            body=[
                export_path,
                {"test.interface.first": {"test_prop": Variant("i", 42)}},
            ],
        )

    # add second interface
    async with ExpectMessage(
        bus1, bus2, "org.freedesktop.DBus.ObjectManager"
    ) as expected_signal:
        bus1.export(export_path, second_interface)
        assert_signal_ok(
            signal=await expected_signal,
            export_path=export_path,
            member="InterfacesAdded",
            signature="oa{sa{sv}}",
            body=[
                export_path,
                {
                    "test.interface.second": {
                        "str_prop": Variant("s", "abc"),
                        "list_prop": Variant("ai", [1, 2, 3]),
                    }
                },
            ],
        )

    # remove single interface
    async with ExpectMessage(
        bus1, bus2, "org.freedesktop.DBus.ObjectManager"
    ) as expected_signal:
        bus1.unexport(export_path, second_interface)
        assert_signal_ok(
            signal=await expected_signal,
            export_path=export_path,
            member="InterfacesRemoved",
            signature="oas",
            body=[export_path, ["test.interface.second"]],
        )

    # add second interface again
    async with ExpectMessage(
        bus1, bus2, "org.freedesktop.DBus.ObjectManager"
    ) as expected_signal:
        bus1.export(export_path, second_interface)
        await expected_signal

    # remove multiple interfaces
    async with ExpectMessage(
        bus1, bus2, "org.freedesktop.DBus.ObjectManager"
    ) as expected_signal:
        bus1.unexport(export_path)
        assert_signal_ok(
            signal=await expected_signal,
            export_path=export_path,
            member="InterfacesRemoved",
            signature="oas",
            body=[export_path, ["test.interface.first", "test.interface.second"]],
        )

    bus1.disconnect()
    bus2.disconnect()
    await asyncio.wait_for(bus1.wait_for_disconnect(), timeout=1)
    await asyncio.wait_for(bus2.wait_for_disconnect(), timeout=1)



import asyncio

import pytest

from dbus_fast.aio import MessageBus
from dbus_fast.annotations import DBusStr
from dbus_fast.service import ServiceInterface, dbus_method


@pytest.mark.asyncio
async def test_multiple_flags_in_message():
    class ExampleInterface(ServiceInterface):
        def __init__(self, name: str) -> None:
            super().__init__(name)

        @dbus_method()
        def Echo(self, what: DBusStr) -> DBusStr:
            return what

    bus = await MessageBus().connect()
    interface = ExampleInterface("test.interface")
    bus.export("/test/path", interface)
    await bus.request_name("test.name")
    bus.disconnect()
    await asyncio.wait_for(bus.wait_for_disconnect(), timeout=1)



import asyncio
import sys

import pytest

from dbus_fast import (
    Message,
    MessageType,
    NameFlag,
    ReleaseNameReply,
    RequestNameReply,
    aio,
    glib,
)
from tests.util import check_gi_repository, skip_reason_no_gi

has_gi = check_gi_repository()


@pytest.mark.asyncio
async def test_name_requests():
    test_name = "aio.test.request.name"

    bus1 = await aio.MessageBus().connect()
    bus2 = await aio.MessageBus().connect()

    async def get_name_owner(name):
        reply = await bus1.call(
            Message(
                destination="org.freedesktop.DBus",
                path="/org/freedesktop/DBus",
                interface="org.freedesktop.DBus",
                member="GetNameOwner",
                signature="s",
                body=[name],
            )
        )

        assert reply.message_type == MessageType.METHOD_RETURN
        return reply.body[0]

    reply = await bus1.request_name(test_name)
    assert reply == RequestNameReply.PRIMARY_OWNER
    reply = await bus1.request_name(test_name)
    assert reply == RequestNameReply.ALREADY_OWNER

    reply = await bus2.request_name(test_name, NameFlag.ALLOW_REPLACEMENT)
    assert reply == RequestNameReply.IN_QUEUE

    reply = await bus1.release_name(test_name)
    assert reply == ReleaseNameReply.RELEASED

    reply = await bus1.release_name("name.doesnt.exist")
    assert reply == ReleaseNameReply.NON_EXISTENT

    reply = await bus1.release_name(test_name)
    assert reply == ReleaseNameReply.NOT_OWNER

    new_owner = await get_name_owner(test_name)
    assert new_owner == bus2.unique_name

    reply = await bus1.request_name(test_name, NameFlag.DO_NOT_QUEUE)
    assert reply == RequestNameReply.EXISTS

    reply = await bus1.request_name(
        test_name, NameFlag.DO_NOT_QUEUE | NameFlag.REPLACE_EXISTING
    )
    assert reply == RequestNameReply.PRIMARY_OWNER

    bus1.disconnect()
    bus2.disconnect()
    await asyncio.wait_for(bus1.wait_for_disconnect(), timeout=1)
    await asyncio.wait_for(bus2.wait_for_disconnect(), timeout=1)


@pytest.mark.skipif(
    sys.version_info[:3][1] in (10, 11, 12, 13),
    reason="segfaults on py3.10,py3.11,py3.12,py3.13",
)
@pytest.mark.skipif(not has_gi, reason=skip_reason_no_gi)
def test_request_name_glib():
    test_name = "glib.test.request.name"
    bus = glib.MessageBus().connect_sync()

    reply = bus.request_name_sync(test_name)
    assert reply == RequestNameReply.PRIMARY_OWNER

    reply = bus.release_name_sync(test_name)
    assert reply == ReleaseNameReply.RELEASED

    bus.disconnect()
